from ctypes import byref, c_double, c_int, pointer

import MCEq.geometry.nrlmsise00 as cmsis
from MCEq.geometry.atmosphere_parameters import (
    DAY_TIMES_SEC,
    DEFAULT_AP,
    DEFAULT_F107,
    DEFAULT_F107A,
    LOCATIONS,
    MONTH_TO_DAY_OF_YEAR,
)
from MCEq.misc import info


class NRLMSISE00Base:
    def __init__(self):
        # Cache altitude value of last call
        self.last_alt = None

        self.inp = cmsis.nrlmsise_input()
        self.output = cmsis.nrlmsise_output()
        self.flags = cmsis.nrlmsise_flags()
        # Use imported constants
        self.month2doy = MONTH_TO_DAY_OF_YEAR
        self.locations = LOCATIONS
        self.daytimes = DAY_TIMES_SEC

        self.current_location = "SouthPole"
        self.init_default_values()

    def surface_vert_depth(self, loc="SouthPole", month="June"):
        self.set_location("SouthPole")
        self.set_season("June")

    def height2depth(self, altitude_cm):
        from scipy.integrate import quad

        return quad(self.get_density, altitude_cm, 112.8 * 1e5, epsrel=0.001)[0]

    def _retrieve_result(self, *args, **kwargs):
        """Calls NRLMSISE library's main function"""
        raise Exception("Not implemented for the base class")

    def get_temperature(self, altitude_cm):
        """Returns temperature in K"""
        self._retrieve_result(altitude_cm)
        return self.output.t[1]

    def get_density(self, altitude_cm):
        """Returns density in g/cm^3"""
        self._retrieve_result(altitude_cm)
        return self.output.d[5]


class cNRLMSISE00(NRLMSISE00Base):
    def init_default_values(self):
        """Sets default to June at South Pole"""

        self.inp.doy = c_int(self.month2doy["June"])  # Day of year
        self.inp.year = c_int(0)  # No effect
        self.inp.sec = c_double(self.daytimes["day"])  # 12:00
        self.inp.alt = c_double(self.locations[self.current_location][2])
        self.inp.g_lat = c_double(self.locations[self.current_location][1])
        self.inp.g_long = c_double(self.locations[self.current_location][0])
        self.inp.lst = c_double(
            self.inp.sec.value / 3600.0 + self.inp.g_long.value / 15.0
        )
        # Do not touch this except you know what you are doing
        # Use imported constants
        self.inp.f107A = c_double(DEFAULT_F107A)
        self.inp.f107 = c_double(DEFAULT_F107)
        self.inp.ap = c_double(DEFAULT_AP)
        self.inp.ap_a = pointer(cmsis.ap_array())
        self.alt_surface = self.locations[self.current_location][2]

        self.flags.switches[0] = c_int(0)
        for i in range(1, 24):
            self.flags.switches[i] = c_int(1)

    def set_location(self, tag):
        if tag not in list(self.locations):
            raise Exception(
                f"NRLMSISE00::set_location(): Unknown location tag '{tag}'."
            )

        self.inp.alt = c_double(self.locations[tag][2])
        self.set_location_coord(*self.locations[tag][:2])
        self.current_location = tag
        self.alt_surface = self.locations[self.current_location][2]

    def set_location_coord(self, longitude, latitude):
        info(5, f"long={longitude:5.2f}, lat={latitude:5.2f}")
        if abs(latitude) > 90 or abs(longitude) > 180:
            raise Exception("NRLMSISE00::set_location_coord(): Invalid inp.")
        self.inp.g_lat = c_double(latitude)
        self.inp.g_long = c_double(longitude)

    def set_season(self, tag):
        if tag not in self.month2doy:
            raise Exception("NRLMSISE00::set_location(): Unknown season tag.")
        info(5, "Season", tag, "doy=", self.month2doy[tag])
        self.inp.doy = self.month2doy[tag]

    def set_doy(self, doy):
        if doy < 0 or doy > 365:
            raise Exception("NRLMSISE00::set_doy(): Day of year out of range.")
        info(5, "day of year", doy)
        self.inp.doy = c_int(doy)

    def _retrieve_result(self, altitude_cm):
        if self.last_alt == altitude_cm:
            return

        inp = self.inp
        inp.alt = c_double(altitude_cm / 1e5)
        cmsis.msis.gtd7_py(
            inp.year,
            inp.doy,
            inp.sec,
            inp.alt,
            inp.g_lat,
            inp.g_long,
            inp.lst,
            inp.f107A,
            inp.f107,
            inp.ap,
            inp.ap_a,
            byref(self.flags),
            byref(self.output),
        )

        self.last_alt = altitude_cm


def test():
    import matplotlib.pyplot as plt
    import numpy as np

    msis = cNRLMSISE00()
    den = np.vectorize(msis.get_density)

    plt.figure(figsize=(16, 5))
    plt.suptitle("NRLMSISE-00")

    h_vec = np.linspace(0, 112.8 * 1e5, 500)
    msis.set_season("January")
    msis.set_location("SouthPole")
    den_sp_jan = den(h_vec)

    msis.set_season("January")
    msis.set_location("Karlsruhe")
    den_ka_jan = den(h_vec)

    plt.subplot(131)
    plt.semilogy(h_vec / 1e5, den_sp_jan, label="MSIS00: SP Jan.")
    plt.semilogy(h_vec / 1e5, den_ka_jan, label="MSIS00: KA Jan.")
    plt.legend()
    plt.xlabel("vertical height in km")
    plt.ylabel(r"density $\rho(h)$ in g/cm$^3$")

    plt.subplot(132)
    plt.plot(h_vec / 1e5, den_ka_jan / den_sp_jan, label="MSIS00: KA/SP")
    plt.xlabel("vertical height in km")
    plt.ylabel(r"density ratio")
    plt.legend(loc="upper left")

    plt.subplot(133)
    msis.set_location("SouthPole")
    for i in range(360 / 30):
        msis.inp.doy = i * 30
        plt.plot(h_vec / 1e5, den(h_vec) / den_sp_jan, label=str(i + 1))
    plt.legend(ncol=2, loc=3)
    plt.title("MSIS00: SouthPole")
    plt.xlabel("vertical height in km")
    plt.ylabel(r"$\rho$(Month) / $\rho$(January)")
    plt.ylim(ymin=0.6)
    plt.tight_layout()

    plt.figure(figsize=(6, 5))
    h2d = np.vectorize(msis.height2depth)
    plt.semilogy(h_vec / 1e5, h2d(h_vec))
    plt.ylabel(r"Slant depth X [g/cm$^2$]")
    plt.xlabel(r"Atmospheric height $h$ [km]")
    plt.subplots_adjust(left=0.15, bottom=0.11)
    plt.show()


if __name__ == "__main__":
    test()



import numpy as np
import pytest
from scipy.interpolate import splrep

from MCEq.particlemanager import _pdata

# DDMEntry


def test_DDMEntry_knot_sigma(ddm_entry):
    expected_result = np.array([0.1, 0.1, 0.1])

    assert np.allclose(ddm_entry.knot_sigma, expected_result)


def test_DDMEntry_fl_ebeam(ddm_entry):
    assert ddm_entry.fl_ebeam == 2.0


def test_DDMEntry_n_knots(ddm_entry):
    assert ddm_entry.n_knots == 3


def test_DDMEntry_x_min(ddm_entry):
    assert ddm_entry.x_min == _pdata.mass(211) / 2


# DDMChannel
def test_DDMChannel_entries(ddm_channel):
    assert len(ddm_channel.entries) == 1


def test_DDMChannel_total_n_knots(ddm_channel):
    assert ddm_channel.total_n_knots == 3


def test_DDMChannel_n_splines(ddm_channel):
    assert ddm_channel.n_splines == 1


def test_DDMChannel_spline_indices(ddm_channel):
    assert ddm_channel.spline_indices == [0]


def test_DDMChannel_get_entry_by_ebeam(ddm_channel):
    entry = ddm_channel.get_entry(ebeam=2.0)
    assert entry.projectile == 2212
    assert entry.secondary == 211


def test_DDMChannel_get_entry_by_idx(ddm_channel):
    entry = ddm_channel.get_entry(idx=0)
    assert entry.projectile == 2212
    assert entry.secondary == 211


def test_DDMChannel_get_entry_raises_error(ddm_channel):
    with pytest.raises(ValueError):
        ddm_channel.get_entry(ebeam=3.0)


def test_DDMChannel_str_representation(ddm_channel):
    expected_output = (
        "\t2212 -> 211:\n\t\t0: ebeam = 2.0 GeV, x17=False, tune v|e=1.000|1.000\n"
    )
    assert str(ddm_channel) == expected_output


# DDMSplineDB


def test_DDMSplineDB_clone_entry(ddm_spline_db):
    ddm_spline_db.clone_entry(2212, 211, 158.0, 1000)
    entry = ddm_spline_db.get_entry(2212, 211, ebeam=1000)
    assert entry.projectile == 2212
    assert entry.secondary == 211


def test_DDMSplineDB_get_spline_indices(ddm_spline_db):
    indices = ddm_spline_db.get_spline_indices(2212, 211)
    assert indices == [0, 1]


def test_DDMSplineDB_channels(ddm_spline_db):
    channels = list(ddm_spline_db.channels)
    assert len(channels) == 1
    channel = channels[0]
    assert channel.projectile == 2212
    assert channel.secondary == 211


def test_DDMSplineDB_get_entry_by_ebeam(ddm_spline_db):
    from MCEq import ddm_utils

    entry = ddm_spline_db.get_entry(2212, 211, ebeam=158)
    assert entry.projectile == 2212
    assert entry.secondary == 211
    assert entry.ebeam == ddm_utils.fmteb(158.0)


def test_DDMSplineDB_get_entry_by_idx(ddm_spline_db):
    from MCEq import ddm_utils

    entry = ddm_spline_db.get_entry(2212, 211, idx=0)
    assert entry.projectile == 2212
    assert entry.secondary == 211
    assert entry.ebeam == ddm_utils.fmteb(31.0)


def test_DDMSplineDB_get_entry_raises_error(ddm_spline_db):
    with pytest.raises(ValueError):
        ddm_spline_db.get_entry(2212, 211, ebeam=123.0)


def test_DDMSplineDB_mk_channel(ddm_spline_db):
    channel = ddm_spline_db._mk_channel(2212, 211)
    assert channel == "2212-211"


# DataDrivenModel


def test_DataDrivenModel_apply_tuning(data_driven_model):
    data_driven_model.apply_tuning(2212, 211, ebeam=158.0, tv=0.5, te=0.8)
    entry = data_driven_model.spline_db.get_entry(2212, 211, ebeam=158.0)
    assert entry.tv == 0.5
    assert entry.te == 0.8


def test_DataDrivenModel_dn_dxl(data_driven_model):
    x = np.linspace(0, 1, 100)
    dn_dxl, error = data_driven_model.dn_dxl(x, 2212, 211, 158.0, return_error=True)
    assert len(dn_dxl) == len(x)
    assert len(error) == len(x)
    assert dn_dxl[0] == 0
    assert error[0] == 0


def test_DataDrivenModel_repr(data_driven_model):
    expected_repr = (
        "DDM channels:\n\t2212 -> 211:\n\t\t0: ebeam = 31.0 GeV, x17=False, "
        + "tune v|e=1.000|1.000\n\t\t1: ebeam = 158.0 GeV, x17=False,"
        + " tune v|e=1.000|1.000\n\n"
    )
    assert repr(data_driven_model) == expected_repr


# ddm_utils


def test_ddm_utils_fmteb():
    from MCEq import ddm_utils

    ebeam_float = 2.5
    ebeam_str = "2.5"
    ebeam_int = 3
    formatted_float = ddm_utils.fmteb(ebeam_float)
    formatted_str = ddm_utils.fmteb(ebeam_str)
    formatted_int = ddm_utils.fmteb(ebeam_int)

    formatted_float == "2.5"
    formatted_str == "2.5"
    formatted_int == "3.0"


def test_ddm_utils_spline_min_max_at_knot():
    from MCEq import ddm_utils

    tck = (
        np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
        3,
    )
    sigma = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
    iknot = 2
    tck_min, tck_max, h = ddm_utils._spline_min_max_at_knot(tck, iknot, sigma)
    expected_tck_min = (np.array([0, 1, 2, 3, 4]), np.array([1, 2, 2.7, 4, 5]), 3)
    expected_tck_max = (np.array([0, 1, 2, 3, 4]), np.array([1, 2, 3.3, 4, 5]), 3)
    expected_h = 0.3

    assert tck_min[0] == pytest.approx(expected_tck_min[0])
    assert tck_min[1] == pytest.approx(expected_tck_min[1])
    assert tck_min[2] == expected_tck_min[2]
    assert tck_max[0] == pytest.approx(expected_tck_max[0])
    assert tck_max[1] == pytest.approx(expected_tck_max[1])
    assert tck_max[2] == expected_tck_max[2]
    assert h == expected_h


def test_generate_DDM_matrix(mceq_qgs, data_driven_model):
    from MCEq import ddm_utils

    channel = data_driven_model.find_channel(2212, 211)

    generated_matrix = ddm_utils._generate_DDM_matrix(
        channel, mceq_qgs, e_min=20, e_max=50, average=True
    )

    expected_matrix = np.array(
        [
            [
                8.00000000e-06,
                4.70861343e-04,
            ],
            [
                0.00000000e00,
                4.00016770e-06,
            ],
        ]
    )

    assert generated_matrix == pytest.approx(expected_matrix)


def test_ddm_utils_eval_spline():
    from MCEq import ddm_utils

    x = np.linspace(0.1, 0.99, 10)
    x_fit = np.linspace(0.1, 0.99, 20)
    y_fit = np.exp(-x_fit)
    tck = splrep(x_fit, y_fit, k=3)
    cov = np.eye(len(tck[1]))

    x17 = True

    # Test without error
    res = ddm_utils._eval_spline(x, tck, x17, cov)
    assert res.shape == (10,)
    assert np.all(res >= 0)

    # Test with error
    res, err = ddm_utils._eval_spline(x, tck, x17, cov, return_error=True)
    assert res.shape == (10,)
    assert err.shape == (10,)
    assert np.all(res >= 0)
    assert np.all(err >= 0)

    # Test with gamma_zfac
    gamma_zfac = 0.5
    res = ddm_utils._eval_spline(x, tck, x17, cov, gamma_zfac=gamma_zfac)
    assert res.shape == (10,)
    assert np.all(res >= 0)


def test_gen_dndx(data_driven_model):
    from MCEq import ddm_utils

    xbins = np.linspace(0.1, 1, 11)
    entry = data_driven_model.spline_db.get_entry(2212, 211, 31)
    dndx = ddm_utils._gen_dndx(xbins, entry)

    # Check the shape of the output
    assert len(dndx) == len(xbins) - 1

    # Check that dndx values are non-negative
    assert np.all(dndx >= 0)

    # Check that dndx values are zero where x < entry.x_min
    assert np.all(dndx[xbins[:-1] < entry.x_min] == 0)


def test_gen_averaged_dndx(data_driven_model):
    from MCEq import ddm_utils

    xbins = np.linspace(0.1, 1, 11)
    entry = data_driven_model.spline_db.get_entry(2212, 211, 31)
    averaged_dndx = ddm_utils._gen_averaged_dndx(xbins, entry)

    # Check the shape of the output
    assert averaged_dndx.shape == xbins[:-1].shape

    # Check that averaged dndx values are non-negative
    assert np.all(averaged_dndx >= 0)

    # Check that averaged dndx values are zero where x < entry.x_min
    assert np.all(averaged_dndx[xbins[:-1] < entry.x_min] == 0)


def test_calc_zfactor_and_error(data_driven_model):
    projectile = 2212
    secondary = 211
    ebeam = 158.0

    entry = data_driven_model.spline_db.get_entry(projectile, secondary, ebeam=ebeam)

    z_factor, z_error = entry.calc_zfactor_and_error()

    # Check that Z-factor values are non-negative
    assert np.all(z_factor >= 0)

    # Check zfactor result
    assert np.allclose(0.0485328868, z_factor)

    # Check zerror result
    assert np.allclose(0.0081361078, z_error)


def test_calc_zfactor_and_error2(data_driven_model):
    projectile = 2212
    secondary = 211
    ebeam = 158.0

    entry = data_driven_model.spline_db.get_entry(projectile, secondary, ebeam=ebeam)

    z_factor, z_error = entry.calc_zfactor_and_error2()

    # Check that Z-factor values are non-negative
    assert np.all(z_factor >= 0)

    # Check zfactor result
    assert np.allclose(0.0590098529, z_factor)

    # Check zerror result
    assert np.allclose(0.010476966, z_error)



import numpy as np


def test_corsika_atmosphere_param_shapes():
    from MCEq.geometry.atmosphere_parameters import _cosika_atmosphere_params

    required_keys = {"_aatm", "_batm", "_catm", "_thickl", "_hlay"}

    for key, param in _cosika_atmosphere_params.items():
        # key is (location, season)
        assert isinstance(key, tuple)
        assert len(key) == 2
        assert isinstance(param, dict)

        # must contain all required keys
        assert set(param.keys()) == required_keys

        # each param must be an ndarray of shape (5,)
        for subkey in required_keys:
            val = param[subkey]
            assert isinstance(val, np.ndarray), f"{key} → {subkey} is not array"
            assert val.shape == (5,), f"{key} → {subkey} has shape {val.shape}"


def test_month_to_day_of_year():
    from MCEq.geometry.atmosphere_parameters import MONTH_TO_DAY_OF_YEAR

    assert len(MONTH_TO_DAY_OF_YEAR) == 12

    # Basic month sanity
    assert MONTH_TO_DAY_OF_YEAR["January"] == 1
    assert MONTH_TO_DAY_OF_YEAR["February"] > MONTH_TO_DAY_OF_YEAR["January"]
    assert MONTH_TO_DAY_OF_YEAR["December"] == 335

    # Monotonic increasing
    vals = list(MONTH_TO_DAY_OF_YEAR.values())
    assert vals == sorted(vals)


def test_locations():
    from MCEq.geometry.atmosphere_parameters import LOCATIONS

    for loc, (lon, lat, h_cm) in LOCATIONS.items():
        assert isinstance(lon, float)
        assert isinstance(lat, float)
        assert isinstance(h_cm, float)
        assert -180.0 <= lon <= 180.0, f"{loc} has invalid longitude"
        assert -90.0 <= lat <= 90.0, f"{loc} has invalid latitude"
        assert h_cm > 0.0, f"{loc} has non-positive altitude"


def test_default_geophysical_constants():
    from MCEq.geometry.atmosphere_parameters import get_nrlmsise00_defaults

    f107a, f107, ap = get_nrlmsise00_defaults()

    assert f107a == 150.0
    assert f107 == 150.0
    assert ap == 4.0


def test_get_location_data():
    from MCEq.geometry.atmosphere_parameters import get_location_data

    loc = get_location_data("SouthPole")
    assert isinstance(loc, tuple)
    assert len(loc) == 3
    lon, lat, h_cm = loc
    assert -180 <= lon <= 180
    assert -90 <= lat <= 90
    assert h_cm > 0

    assert get_location_data("NotARealPlace") is None


def test_get_month_day_of_year():
    from MCEq.geometry.atmosphere_parameters import get_month_day_of_year

    assert get_month_day_of_year("January") == 1
    assert get_month_day_of_year("December") == 335
    assert get_month_day_of_year("NotAMonth") is None


def test_get_day_time_seconds():
    from MCEq.geometry.atmosphere_parameters import get_day_time_seconds

    assert get_day_time_seconds("day") == 43200.0
    assert get_day_time_seconds("night") == 0.0
    assert get_day_time_seconds("dusk") is None



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "MCEq"
copyright = "2026, Anatoli Fedynitch, Stefan Fröse"
author = "Anatoli Fedynitch, Stefan Fröse"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "sphinx_copybutton",
    "sphinx_design",
    "nbsphinx",
    "sphinx_gallery.load_style",
    "matplotlib.sphinxext.plot_directive",
]


templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


source_suffix = ".rst"
master_doc = "index"


intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable", None),
    "crflux": ("https://crfluxmodels.readthedocs.io/en/latest", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

html_title = f"{project}"
htmlhelp_basename = project + "docs"

html_theme_options = {
    "navbar_links": [
        {"name": "Quickstart", "url": "quickstart/index", "internal": True},
        {"name": "Tutorial", "url": "tutorial", "internal": True},
        {"name": "Example Gallery", "url": "example-gallery", "internal": True},
        {"name": "v1.4", "url": "v14v13_diff", "internal": True},
        {"name": "Cite Us", "url": "citeus", "internal": True},
        {"name": "API Reference", "url": "api-reference/index", "internal": True},
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/mceq-project/mceq",
            "icon": "fa-brands fa-github",
        },
    ],
    "announcement": "MCEq v1.4 is now available!",
}

html_sidebars = {
    "quickstart/index": [],  # disable sidebar on Quickstart
    "tutorial": [],
    "citeus": [],  # disable sidebar on Cite Us
    "example-gallery": [],  # disable sidebar on References
    "v12v11_diff": [],  # disable sidebar on References
    "v14v13_diff": [],  # disable sidebar on References
}

html_show_sourcelink = False



from copy import copy
from typing import Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
from jacobi import propagate
from scipy.integrate import quad
from scipy.interpolate import splev

from .misc import info

_LIMIT_PROPAGATE = np.inf
_PROPAGATE_PARAMS = dict(maxiter=10, maxgrad=1, method=0)  # Used to compute the
_QUAD_PARAMS = dict(limit=150, epsabs=1e-5)


def fmteb(ebeam: Union[float, str, int]) -> str:
    """
    Format beam energy for handling in db.

    Parameters
    ----------
    ebeam : Union[float, str, int]
        The beam energy.

    Returns
    -------
    str
        The formatted beam energy.
    """
    return f"{float(ebeam):.1f}"


def _spline_min_max_at_knot(
    tck: Tuple, iknot: int, sigma: npt.NDArray
) -> Tuple[Tuple, Tuple, float]:
    """
    Return variation of spline coefficients by 1 sigma.

    Args:
        tck : Tuple
            A tuple representing the input spline.
        iknot : int
            The knot index to calculate the variation of.
        sigma : npt.NDArray
            The 1-sigma errors of each knot.

    Returns
    -------
    Tuple[Tuple, Tuple, float]
        A tuple containing two tuples representing the minimum and maximum spline
        coefficients, respectively, and a float representing the variation of the
        spline coefficients by 1 sigma.
    """
    assert iknot <= len(tck[1]) - 1, f"Invalid knot {iknot} requested"
    _c_min = np.copy(tck[1])
    _c_max = np.copy(tck[1])
    h = sigma[iknot]
    _c_min[iknot] = tck[1][iknot] - h
    _c_max[iknot] = tck[1][iknot] + h
    _tck_min = (tck[0], _c_min, tck[-1])
    _tck_max = (tck[0], _c_max, tck[-1])
    return _tck_min, _tck_max, h


def _generate_DDM_matrix(
    channel, mceq, e_min: float = -1.0, e_max: float = -1.0, average: bool = True
) -> npt.NDArray:
    """
    Generate a Data-Driven Model matrix for the given channel and MCEq object.

    Args:
        channel : _DDMChannel
            The DDM channel.
        mceq : MCEq
            The MCEq object.
        e_min : float, optional
            The minimum energy range in GeV where DDM cross sections overwrite
            original MCEq matrices, by default -1.0.
        e_max : float, optional
            The maximum energy range in GeV where DDM cross sections overwrite
            original MCEq matrices, by default -1.0.
        average : bool, optional
            Whether to average the dndx values, by default True.

    Returns:
        npt.NDArray
            The DDM matrix.
    """
    from MCEq.misc import _eval_energy_cuts, energy_grid

    projectile = channel.projectile
    secondary = channel.secondary
    # Definitions of particles and energy grid
    e_min = mceq._energy_grid.b[0] if e_min < 0 else e_min
    e_max = mceq._energy_grid.b[-1] if e_max < 0 else e_max
    mceq_egr = mceq._energy_grid
    min_idx, max_idx, _cuts = _eval_energy_cuts(mceq_egr.c, e_min, e_max)
    _energy_grid = energy_grid(
        mceq_egr.c[_cuts],
        mceq_egr.b[min_idx : max_idx + 1],
        mceq_egr.w[_cuts],
        int(max_idx - min_idx),
    )
    dim = len(_energy_grid.c)

    projectile_mass = mceq.pman[projectile].mass
    secondary_mass = mceq.pman[secondary].mass

    # Convert from kinetic to lab energy
    elab_secondary_bins = _energy_grid.b + secondary_mass
    elab_secondary_centers = _energy_grid.c + secondary_mass
    elab_proj_centers = _energy_grid.c + projectile_mass
    e_widths = np.diff(elab_secondary_bins)
    xgrid = elab_secondary_bins / elab_secondary_centers[-1]

    # Pull a copy of the original matrix
    try:
        # If matrices are stored on GPU, copy them to CPU
        mat = copy(mceq.pman[projectile].hadr_yields[mceq.pman[secondary]].get())
    except AttributeError:
        mat = copy(mceq.pman[projectile].hadr_yields[mceq.pman[secondary]])
    mat = mat[min_idx : max_idx + 1, min_idx : max_idx + 1]
    if average:
        dndx_generator = _gen_averaged_dndx
    else:
        dndx_generator = _gen_dndx

    info(
        5,
        f"DDM will be used between {e_min:.1e} GeV and {e_max:.1e} GeV",
    )
    n_datasets = channel.n_splines
    assert n_datasets > 0, "Number of splines per channel can't be zero"

    if n_datasets == 1:
        entry = channel.get_entry(idx=0)
        ie0 = np.argmin(np.abs(entry.fl_ebeam - elab_proj_centers))
        info(
            5,
            f"Dataset 0 ({projectile}, {secondary}): mceq_eidx={ie0}, "
            + "mceq_ebeam={mceq_ebeam:4.3f}, ebeam={entry.ebeam}",
        )
        averaged_dndx = dndx_generator(xgrid, entry)

        for ie, eproj in enumerate(elab_proj_centers[:-1]):
            if (eproj < e_min) or (eproj > e_max):
                info(0, "Skipping out of range energy", eproj)
                continue

            mat[: ie + 1, ie] = (
                averaged_dndx[dim - ie - 1 :] / elab_proj_centers[ie] * e_widths[ie]
            )
    else:
        # Loop through datasets in pairs of two and interpolate between them
        for interval in range(n_datasets - 1):
            entry_0 = channel.get_entry(idx=interval)
            entry_1 = channel.get_entry(idx=interval + 1)
            ie0 = np.argmin(np.abs(entry_0.fl_ebeam - elab_proj_centers))
            ie1 = np.argmin(np.abs(entry_1.fl_ebeam - elab_proj_centers))
            mceq_ebeam_0 = elab_proj_centers[ie0]
            mceq_ebeam_1 = elab_proj_centers[ie1]
            info(
                5,
                f"Dataset {interval} ({projectile}, {secondary}): "
                + f"mceq_eidx={ie0}, mceq_ebeam={mceq_ebeam_0:4.3f}, "
                + f"ebeam={entry_0.ebeam}",
            )
            info(
                5,
                f"Dataset {interval + 1} ({projectile}, {secondary}): "
                + f"mceq_eidx={ie1}, mceq_ebeam={mceq_ebeam_1:4.3f}, "
                + f"ebeam={entry_1.ebeam}",
            )

            try:
                averaged_dndx_0 = dndx_generator(xgrid, entry_0)
            except ValueError:
                raise Exception(
                    "Error averaging cross section for "
                    + f"({projectile},{secondary},{entry_0.ebeam})"
                )

            try:
                averaged_dndx_1 = dndx_generator(xgrid, entry_1)
            except ValueError:
                raise Exception(
                    "Error averaging cross section for "
                    + f"({projectile},{secondary},{entry_1.ebeam})"
                )

            for ie, eproj in enumerate(elab_proj_centers):
                if (eproj < e_min) or (eproj > e_max):
                    continue
                if ie <= ie0 and interval == 0:
                    # If it is the first interval, extrapolate from the lowest energies
                    mat[: ie + 1, ie] = (
                        averaged_dndx_0[dim - ie - 1 :]
                        / elab_proj_centers[ie]
                        * e_widths[ie]
                    )
                elif ie < ie0 and interval > 0:
                    # if it is not the first interval, don't touch the columns
                    # that have been previously filled
                    continue
                elif ie <= ie1:
                    # Interpolate between two datasets
                    f0 = (
                        averaged_dndx_0[dim - ie - 1 :]
                        / elab_proj_centers[ie]
                        * e_widths[ie]
                    )
                    f1 = (
                        averaged_dndx_1[dim - ie - 1 :]
                        / elab_proj_centers[ie]
                        * e_widths[ie]
                    )
                    m = (f1 - f0) / np.log(entry_1.fl_ebeam / entry_0.fl_ebeam)
                    mat[: ie + 1, ie] = f0 + m * np.log(eproj / entry_0.fl_ebeam)
                elif ie > ie1 and interval != n_datasets - 2:
                    # if it is not the first dataset, don't touch the columns
                    # that have been previously filled
                    break
                else:
                    mat[: ie + 1, ie] = (
                        averaged_dndx_1[dim - ie - 1 :]
                        / elab_proj_centers[ie]
                        * e_widths[ie]
                    )

    return mat


def _eval_spline(
    x: Union[float, npt.NDArray],
    tck: Tuple[npt.NDArray, npt.NDArray, int],
    x17: bool,
    cov: npt.NDArray,
    return_error: Optional[bool] = False,
    gamma_zfac: Optional[float] = None,
) -> Union[npt.NDArray, Tuple[npt.NDArray, npt.NDArray]]:
    """
    Evaluate the spline in dn/dx.

    Parameters
    ----------
    x : float or numpy.ndarray
        The knot points.
    tck : Tuple[numpy.ndarray, numpy.ndarray, int]
        A tuple containing the knots of the spline.
    x17 : bool
        If True, scale by x^1.7 dn/dx. Otherwise, return only dn/dx.
    cov : numpy.ndarray
        The covariance matrix of the spline.
    return_error : bool, optional
        If True, return the propagated error of the fit. Default is False.
    gamma_zfac : float, optional
        The exponent for the gamma function. If None, it is not used. Default is None.

    Returns
    -------
    numpy.ndarray or Tuple[numpy.ndarray, numpy.ndarray]
        The value of the fit at each knot point. If `return_error` is True,
        it also returns the propagated error of the fit.
    """
    if gamma_zfac is None:
        factor = x ** (-1.7) if x17 else 1.0
    else:
        factor = x ** (gamma_zfac) if not x17 else x ** (gamma_zfac - 1.7)

    def func(tck_1: npt.NDArray) -> npt.NDArray:
        return factor * np.exp(splev(x, (tck[0], tck_1, tck[2])))

    func_params = tck[1]

    if return_error:
        y, C = propagate(func, func_params, cov, **_PROPAGATE_PARAMS)
        sig_y = np.squeeze(np.sqrt(np.diag(np.atleast_1d(C))))
        return y, sig_y
    else:
        res = np.atleast_1d(func(func_params))
        res[(res < 0) | ~np.isfinite(res) | (res > _LIMIT_PROPAGATE)] = 0.0
        return res.squeeze()


def _eval_spline_and_correction(
    x: Union[float, npt.NDArray],
    tck_dndx: Tuple[npt.NDArray, npt.NDArray, int],
    tck_corr: Tuple[npt.NDArray, npt.NDArray, int],
    x17: bool,
    cov: npt.NDArray,
    return_error: Optional[bool] = False,
    gamma_zfac: Optional[float] = None,
) -> Union[npt.NDArray, Tuple[npt.NDArray, npt.NDArray]]:
    """
    This case evaluates the product of two splines, which has been
    used in DDM to apply the MC correction from pp -> K+ to pC -> K+

    Args:
        x: The knot points.
        tck_dndx: A tuple containing the knots of the spline.
        tck_corr: Knots of the spline that is multiplied with the dndx spline.
        x17: If True, fit x^1.7 dn/dx. Otherwise, fit dn/dx.
        cov: The covariance matrix .
        tv: The tuning value.
        te: The error scale.
        return_error: If True, return the propagated error of the fit.
        gamma_zfac: The exponent for the gamma function. If None, it is not used.

    Returns:
        The value of the fit at each knot point.
    """
    if gamma_zfac is None:
        factor = x ** (-1.7) if x17 else 1.0
    else:
        factor = x ** (gamma_zfac) if not x17 else x ** (gamma_zfac - 1.7)

    func_params = np.hstack([tck_dndx[1], tck_corr[1]])

    def func(params: npt.NDArray) -> npt.NDArray:
        tck_dndx_1 = params[: len(tck_dndx[1])]
        tck_corr_1 = params[len(tck_dndx[1]) :]
        return (
            factor
            * np.exp(splev(x, (tck_dndx[0], tck_dndx_1, tck_dndx[2])))
            * splev(x, (tck_corr[0], tck_corr_1, tck_corr[2]))
        )

    if return_error:
        y, C = propagate(func, func_params, cov, **_PROPAGATE_PARAMS)
        sig_y = np.squeeze(np.sqrt(np.diag(np.atleast_1d(C))))
        return y, sig_y
    else:
        res = np.atleast_1d(func(func_params))
        res[(res < 0) | ~np.isfinite(res) | (res > _LIMIT_PROPAGATE)] = 0.0
        return res.squeeze()


def _gen_dndx(xbins: np.ndarray, entry) -> np.ndarray:
    """
    Generate dN/dx values for a given energy binning using a DDM entry.

    Parameters
    ----------
    xbins : numpy.ndarray
        Binning of x values.
    entry : _DDMEntry
        The DDM entry containing the spline information.

    Returns
    -------
    numpy.ndarray
        Array of dN/dx values corresponding to the provided xbinning.

    Notes
    -----
    The dN/dx values are calculated by evaluating the spline defined by the
    DDM entry at the square root of the bin boundaries. The resulting values
    are multiplied by the tuning value of the entry.

    If the evaluated x value is less than the minimum x value of the spline,
    the corresponding dN/dx value is set to zero.

    The returned array has the same shape as the provided xbinning.

    Examples
    --------
    >>> xbins = np.linspace(0, 1, 11)
    >>> entry = _DDMEntry(...)
    >>> dndx = _gen_dndx(xbins, entry)
    """

    x = np.sqrt(xbins[1:] * xbins[:-1])

    res = _eval_spline(x, entry.tck, entry.x17, entry.cov, return_error=False)
    res[x < entry.x_min] = 0
    return entry.tv * res


def _gen_averaged_dndx(xbins: np.ndarray, entry) -> np.ndarray:
    """
    Generate averaged dN/dx values for a given energy binning using a DDM entry.

    Parameters
    ----------
    xbins : numpy.ndarray
        Binning of x values.
    entry : _DDMEntry
        The DDM entry containing the spline information.

    Returns
    -------
    numpy.ndarray
        Array of averaged dN/dx values corresponding to the provided xbinning.

    Notes
    -----
    The averaged dN/dx values are calculated by integrating the spline defined
    by the DDM entry over each energy bin and dividing by the bin width. The
    resulting values are multiplied by the tuning value of the entry.

    If an energy bin extends beyond x=1, the integration is performed from the
    lower limit of the bin to x=1.

    If an energy bin is entirely below entry.x_min, the averaged dN/dx value
    for that bin is set to zero.

    The returned array has the same shape as the provided xbinning[:-1].

    Examples
    --------
    >>> xbins = np.linspace(0.1, 1, 11)
    >>> entry = _DDMEntry(...)
    >>> averaged_dndx = _gen_averaged_dndx(xbins, entry)
    """

    xwidths = np.diff(xbins)
    integral = np.zeros_like(xwidths)

    for ib in range(len(integral)):
        if xbins[ib + 1] > 1:
            integral[ib] = entry.tv * (
                quad(
                    _eval_spline,
                    xbins[ib],
                    1.0,
                    args=(entry.tck, entry.x17, entry.cov),
                    **_QUAD_PARAMS,
                )[0]
                / xwidths[ib]
            )
        else:
            if xbins[ib + 1] < entry.x_min:
                integral[ib] = 0.0
                continue
            elif xbins[ib] < entry.x_min:
                low_lim = entry.x_min
            else:
                low_lim = xbins[ib]

            integral[ib] = entry.tv * (
                quad(
                    _eval_spline,
                    low_lim,
                    xbins[ib + 1],
                    args=(entry.tck, entry.x17, entry.cov),
                    **_QUAD_PARAMS,
                )[0]
                / xwidths[ib]
            )

    return integral


def gen_matrix_variations(ddm_obj, mceq):
    """
    Generate variations of DDM matrices and isospin partners.

    Parameters
    ----------
    ddm_obj : DataDrivenModel
        DataDrivenModel object.
    mceq : MCEq
        MCEq object.

    Returns
    -------
    Dict[Tuple[int, int], Dict[int, List[numpy.ndarray]]]
        Dictionary mapping (projectile, secondary) pairs to a dictionary
        of variations for each knot index. Each variation consists of a
        list of two matrices (max and min) and the variation magnitude.
    Dict[Tuple[int, int], Tuple[Tuple[int, int], Dict[int, List[numpy.ndarray]]]]
        Dictionary mapping (projectile, secondary) pairs to their isospin
        partners, along with the variations of isospin matrices.

    Examples
    --------
    >>> model = DataDrivenModel(...)
    >>> mceq = MCEq(...)
    >>> matrix_variations, isospin_partners = gen_matrix_variations(model, mceq)
    """

    matrix_variations = {}
    isospin_partners = {}

    ddm_matrices = ddm_obj.ddm_matrices(mceq)
    for channel in ddm_obj.spline_db.channels:
        info(1, f"Generating variations for channel\n{channel}")
        mat_db = {}
        iso_part_db = {}

        for ispl, spl_entry in enumerate(channel.entries):
            tck = spl_entry.tck
            sig = spl_entry.knot_sigma
            mat_db[ispl] = {}
            iso_part_db[ispl] = {}

            for iknot in range(0, spl_entry.n_knots):
                if np.allclose(tck[1][iknot], 0.0):
                    continue

                tck_min, tck_max, h = _spline_min_max_at_knot(tck, iknot, sig)
                spl_entry.tck = tck_max
                mat_max = _generate_DDM_matrix(
                    channel, mceq, e_min=ddm_obj.e_min, e_max=ddm_obj.e_max
                )
                spl_entry.tck = tck_min
                mat_min = _generate_DDM_matrix(
                    channel, mceq, e_min=ddm_obj.e_min, e_max=ddm_obj.e_max
                )
                spl_entry.tck = tck

                mat_db[ispl][iknot] = [mat_max, mat_min, h]

                if channel.secondary in [321, -321] and ddm_obj.enable_K0_from_isospin:
                    imat_max = 0.5 * (
                        mat_max + ddm_matrices[(channel.projectile, -channel.secondary)]
                    )
                    imat_min = 0.5 * (
                        mat_min + ddm_matrices[(channel.projectile, -channel.secondary)]
                    )
                    iso_part_db[ispl][iknot] = [imat_max, imat_min, h]

        matrix_variations[(channel.projectile, channel.secondary)] = mat_db
        if channel.secondary in [321, -321]:
            isospin_partners[(channel.projectile, channel.secondary)] = (
                (
                    310,
                    130,
                ),
                iso_part_db,
            )

    return matrix_variations, isospin_partners



import pathlib

import crflux.models as pm
import numpy as np
import pytest
from scipy.sparse import csr_matrix

from MCEq import config
from MCEq.core import MCEqRun


@pytest.fixture(scope="session")
def mceq_sib21():
    config.debug_level = 2
    config.cuda_gpu_id = 0

    config.mceq_db_fname = "mceq_db_v140reduced_compact.h5"

    if config.has_mkl:
        config.set_mkl_threads(2)

    return MCEqRun(
        interaction_model="SIBYLL21",
        theta_deg=0.0,
        primary_model=(pm.HillasGaisser2012, "H3a"),
    )


@pytest.fixture(scope="session")
def mceq_qgs():
    config.debug_level = 2
    config.cuda_gpu_id = 0

    config.mceq_db_fname = "mceq_db_v140reduced_compact.h5"

    if config.has_mkl:
        config.set_mkl_threads(2)

    return MCEqRun(
        interaction_model="QGSJETII04",
        theta_deg=0.0,
        primary_model=(pm.HillasGaisser2012, "H3a"),
    )


@pytest.fixture(scope="function")
def ddm_entry():
    from MCEq import ddm, ddm_utils

    entry = ddm._DDMEntry(
        ebeam=ddm_utils.fmteb(2.0),
        projectile=2212,
        secondary=211,
        x17=False,
        tck=(np.array([1, 2, 3]), np.array([4, 5, 6]), 3),
        cov=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        tv=1.0,
        te=0.1,
        spl_idx=1,
    )
    return entry


@pytest.fixture(scope="function")
def ddm_channel():
    from MCEq import ddm, ddm_utils

    ch = ddm._DDMChannel(projectile=2212, secondary=211)
    ch.add_entry(
        ebeam=ddm_utils.fmteb(2.0),
        projectile=2212,
        secondary=211,
        x17=False,
        tck=(np.array([1, 2, 3]), np.array([4, 5, 6]), 3),
        cov=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        tv=1.0,
        te=1.0,
    )

    return ch


@pytest.fixture(scope="function")
def ddm_spline_db():
    from MCEq import ddm

    db = ddm.DDMSplineDB(
        enable_channels=[(2212, 211)],
        exclude_projectiles=[111, 2112],
    )
    return db


@pytest.fixture(scope="function")
def data_driven_model():
    from MCEq import ddm

    _ddm = ddm.DataDrivenModel(
        e_min=5.0,
        e_max=500.0,
        enable_channels=[(2212, 211)],
        exclude_projectiles=[111, 2112],
        enable_K0_from_isospin=True,
    )
    return _ddm


@pytest.fixture
def msis_expected_file(request):
    test_dir = pathlib.Path(request.fspath).parent
    path = test_dir / "msis_expected.txt"
    if not path.exists():
        raise FileNotFoundError(
            f"Expected output file {path} not found. "
            "Please run the test with the expected output file."
        )
    return path


@pytest.fixture(scope="function")
def toy_solver_problem():
    import numpy as np
    from scipy.sparse import csr_matrix

    nsteps = 10
    size = 5
    dX = np.full(nsteps, 0.1)
    rho_inv = np.ones(nsteps)
    grid_idcs = list(range(nsteps))

    # mimic how self.int_m and self.dec_m are used in solve()
    lam_int = 0.3
    lam_dec = 0.1
    int_m = csr_matrix(-lam_int * np.eye(size))  # simple interaction term
    dec_m = csr_matrix(-lam_dec * np.eye(size))  # simple decay term

    phi0 = np.ones(size)
    return nsteps, dX, rho_inv, int_m, dec_m, phi0, grid_idcs


@pytest.fixture(scope="session")
def toy_solver_setup():
    nsteps = 10
    size = 5
    dX = np.full(nsteps, 0.1)
    rho_inv = np.ones(nsteps)
    grid_idcs = list(range(nsteps))

    lam_int = 0.3
    lam_dec = 0.1
    int_m = csr_matrix(-lam_int * np.eye(size))
    dec_m = csr_matrix(-lam_dec * np.eye(size))

    phi = np.ones(size)

    return nsteps, dX, rho_inv, int_m, dec_m, phi, grid_idcs



import importlib
import os
import pathlib
import platform
import sys
import warnings

from MCEq import base_path

#: Debug flag for verbose printing, 0 silences MCEq entirely
debug_level = 1
#: Override debug prinput for functions listed here (just give the name,
#: "get_solution" for instance) Warning, this option slows down initialization
#: by a lot. Use only when needed.
override_debug_fcn = []
#: Override debug printout for debug levels < value for the functions above
override_max_level = 10
#: Print module name in debug output
print_module = False

# =================================================================
# Paths and library locations
# =================================================================

#: Directory where the data files for the calculation are stored
data_dir = pathlib.Path(base_path) / "data"

#: File name of the MCEq database
mceq_db_fname = "mceq_db_lext_dpm193_v140.h5"

#: File name of the MCEq database
em_db_fname = "mceq_db_EM_Tsai-Max_Z7.31.h5"

# =================================================================
# Atmosphere and geometry settings
# =================================================================

#: The latest versions of MCEq work in kinetic energy not total energy
#: If you want the result to be compatible with the previous choose
#: 'total energy' else 'kinetic energy'
return_as = "kinetic energy"
#: Atmospheric model in the format: (model, (arguments))
density_model = ("CORSIKA", ("BK_USStd", None))
#: density_model = ('MSIS00_IC',('SouthPole','January'))
#: density_model = ('GeneralizedTarget', None)

#: Definition of prompt (only for correct accounting). Leptons from parent particles
#: with ctau < prompt_ctau will be counted in the pr_[mu, numu, nue] category, whereas
#: everything else will be attributed to the "conventional" category
# cm (everything shorter-lived than K0s will be considered prompt)
prompt_ctau = 2.6842

#: Approximate value for the maximum density expected. Needed for the
#: resonance approximation. Default value: air at the surface
max_density = 0.001225
#: Material for interaction lengths, ionization and radiation (=continuous) loss terms
#: Currently available choices: 'air', 'water', 'ice', 'rock', 'co2', 'hydrogen', 'iron'
interaction_medium = "air"

#: Average target mass (for interaction length calculations)
#: Change parameter only in combination with interaction model setting.
#: By default, secondary particle production matrices are calculated for air targets
#: If set to 'auto', use default according to the "interaction_medium" settings below
A_target = "auto"

#: parameters for EarthGeometry
r_E = 6391.0e3  # Earth radius in m
h_obs = 0.0  # observation level in m
h_atm = 112.8e3  # top of the atmosphere in m
X_start = 0.0  # starting slant depth in g/cm^-2

#: Default parameters for GeneralizedTarget
#: Total length of the target [m]
len_target = 1000.0
#: density of default material in g/cm^3
env_density = 0.001225
env_name = "air"
#: Approximate value for the maximum density expected. Needed for the
#: resonance approximation. Default value: air at the surface
max_density = (0.001225,)
# =================================================================
# Parameters of numerical integration
# =================================================================

#: Minimal energy for grid
#: The minimal energy (technically) is 1e-2 GeV. Currently you can run into
#: stability problems with the integrator with such low thresholds. Use with
#: care and check results for oscillations and feasibility.
e_min = 0.1

#: The maximal energy is 1e12 GeV, but not all interaction models run at such
#: high energies. If you are interested in lower energies, reduce this value
#: for inclusive calculations to max. energy of interest + 4-5 orders of
#: magnitude. For single primaries the maximal energy is directly limited by
#: this value. Smaller grids speed up the initialization and integration.
e_max = 1e11

#: Enable electromagnetic cascade with matrices from EmCA
enable_em = False

#: Selection of integrator (euler/odepack)
integrator = "euler"

#: euler kernel implementation (numpy/MKL/CUDA/accelerate).
#: With serious nVidia GPUs CUDA a few times faster than MKL
#: autodetection of fastest kernel below
kernel_config = "auto"

#: Select CUDA device ID if you have multiple GPUs
cuda_gpu_id = 0

#: CUDA Floating point precision (default 32-bit 'float')
cuda_fp_precision = 64

#: Floating point precision (is set automatically)
floatlen = None

#: Number of MKL threads (for sparse matrix multiplication the performance
#: advantage from using more than a few threads is limited by memory bandwidth)
#: Irrelevant for GPU integrators, but can affect initialization speed if
#: numpy is linked to MKL.
mkl_threads = 8

#: parameters for the odepack integrator. More details at
#: http://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.ode.html#scipy.integrate.ode
ode_params = {
    "name": "lsoda",
    "method": "bdf",
    "nsteps": 1000,
    # 'max_step': 10.0,
    "rtol": 0.01,
}

# =========================================================================
# Advanced settings
# =========================================================================

#: The leading process is can be either "decays" or "interactions". This depends
#: on the target density and it is usually chosen automatically. For
#: advanced applications one can force "interactions" to be the dominant
#: process. Essentially this affects how the adaptive step size is computed.
#: There is also the choice of "auto" that takes both processes into account
leading_process = "auto"

#: Stability margin for the integrator. The default 0.95 means that step
#: sizes are chosen 5% away from the stability circle. Usually no need to
#: change, except you know what it does.
stability_margin = 0.95

#: Ratio of decay_length/interaction_length where particle interactions
#: are neglected and the resonance approximation is used
#: 0.5 ~ precision loss <+3% speed gain ~ factor 10
#: If smoothness and shape accuracy for prompt flux is crucial, use smaller
#: values around 0.1 or 0.05
hybrid_crossover = 0.5

#: Maximal integration step dX in g/cm2. No limit necessary in most cases,
#: use for debugging purposes when searching for stability issues, especially
#: if e_min is < 1 GeV.
dXmax = 1.0

#: Minimal CR nucleon energy in primary model. If (low energy)
#: hadronic interaction model doesn't properly implement interactions
#: or cross sections, nucleons can "drop through" without cascading

minimal_primary_energy = 3.0

#: Enable default tracking particles, such as pi_numu, pr_mu+, etc.
#: If only total fluxes are of interest, disable this feature to gain
#: performance since the eqution system becomes smaller and sparser
enable_default_tracking = True

#: Ionization and radiative losses according to stopping power tables (PDG)
enable_energy_loss = True

#: Apply stopping power to all charged hadrons (muon dEdX is used and is ~ok)
generic_losses_all_charged = False

#: Treat radiation (bremsstrahlung) as continuous loss, disable if explicit
#: electromagnetic cross sections available
enable_cont_rad_loss = True

#: Fall-back to air production matrices if medium not included in data file
fallback_to_air_cs = True

#: enable EM ionization loss for electrons and positrons
enable_em_ion = True

#: Improve (explicit solver) stability by averaging the continous loss
#: operator
average_loss_operator = True

#: Step size (dX) for averaging
loss_step_for_average = 1e-1

#: Raise exception when requesting unknown particles from get_solution
excpt_on_missing_particle = False

#: When using modified particle production matrices use
#: isospin symmetries to determine the corresponding
#: modification for neutrons and K0L/K0S
use_isospin_sym = True

#: Helicity dependent muons decays from analytical expressions
muon_helicity_dependence = True

#: Assume nucleon, pion and kaon cross sections for interactions of
#: rare or exotic particles (mostly relevant for non-compact mode)
assume_nucleon_interactions_for_exotics = True

#: This is not used in the code as before, instead the low energy
#: extension is compiled into the HDF backend files.
low_energy_extension = {
    "he_le_transition": 80,  # GeV
    "nbins_interp": 3,
    "use_unknown_cs": True,
}

#: Advanced settings (some options might be obsolete/not working)
adv_set = {
    #: Disable particle production by all hadrons, except nucleons
    "disable_interactions_of_unstable": False,
    #: Disable particle production by charm *projectiles* (interactions)
    "disable_charm_pprod": False,
    #: Disable resonance/prompt contribution (this group of options
    #: is either obsolete or needs maintenance.)
    #: "disable_resonance_decay" : False,
    #: Allow only those particles to be projectiles (incl. anti-particles)
    #: Faster initialization,
    #: For inclusive lepton flux computations:
    #: precision loss ~ 1%, for SIBYLL2.3.X with charm 5% above 10^7 GeV
    #: Might be different for yields (set_single_primary_particle)
    #: For full precision or if in doubt, use []
    "allowed_projectiles": [],  # [2212, 2112, 211, 321, 130, 11, 22],
    #: Disable particle (production)
    "disabled_particles": [],  # 20, 19, 18, 17, 97, 98, 99, 101, 102, 103
    #: Disable leptons coming from prompt hadron decays at the vertex
    "disable_direct_leptons": False,
    #: Difficult to explain parameter
    "disable_leading_mesons": False,
    #: Do not apply mixing to these particles
    "exclude_from_mixing": [13],
    #: Switch off decays. E.g., disable muon decay with [13,-13]
    "disable_decays": [],
    #: Force particles to be treated as resonance
    "force_resonance": [],
    #: Disable mixing between resonance approx. and full propagation
    "no_mixing": False,
    #: Force the interaction cross sections to a specific model
    "forced_int_cs": None,
    #: Replace only the meson air cross sections with that from a different model
    "replace_meson_cross_sections_with": None,
}

#: Particles for compact mode
standard_particles = [11, 12, 13, 14, 16, 211, 321, 2212, 2112, 3122, 411, 421, 431]

#: Anti-particles
standard_particles += [-pid for pid in standard_particles]

#: unflavored particles
#: append 221, 223, 333, if eta, omega and phi needed directly
standard_particles += [22, 111, 130, 310]  #: , 221, 223, 333]

#: This construct provides access to the attributes as in previous
#: versions, using `from mceq_config import config`. The future versions
#: will access the module attributes directly.

#: Autodetect best solver
#: determine shared library extension and MKL path
pf = platform.platform()
has_accelerate = False

prefix = pathlib.Path(sys.prefix)
if "Linux" in pf:
    mkl_libs = list((prefix / "lib").glob("libmkl_rt*"))
    mkl_path = mkl_libs[0] if mkl_libs else prefix / "lib" / "libmkl_rt.so"
elif "macOS" in pf:
    mkl_path = prefix / "lib" / "libmkl_rt.dylib"
    has_accelerate = True
else:
    # Windows or unknown OS: search for mkl_rt*.dll in Library/bin and lib
    mkl_path = None
    mkl_dirs = [prefix / "Library" / "bin", prefix / "lib"]
    mkl_candidates = []
    for d in mkl_dirs:
        if d.exists():
            mkl_candidates.extend(d.glob("mkl_rt*.dll"))
    if mkl_candidates:
        mkl_path = mkl_candidates[0]
    else:
        # fallback to default path
        mkl_path = prefix / "Library" / "bin" / "mkl_rt.dll"

    mkl_path = os.fspath(mkl_path)

# mkl library handler
mkl = None

has_mkl = bool(pathlib.Path(mkl_path).is_file())

# Look for cupy module
has_cuda = importlib.util.find_spec("cupy") is not None

# CUDA is usually fastest, then MKL. Fallback to numpy.
if kernel_config == "auto":
    if has_cuda:
        kernel_config = "CUDA"
    elif has_mkl:
        kernel_config = "MKL"
    elif has_accelerate:
        kernel_config = "accelerate"
    else:
        kernel_config = "numpy"
else:
    if kernel_config.lower() == "cuda" and not has_cuda:
        raise Exception("CUDA unavailable. Make sure cupy is installed.")
    elif kernel_config.lower() == "mkl" and not has_mkl:
        raise Exception("MKL unavailable. Make sure Intel MKL is installed.")
    elif kernel_config.lower() == "accelerate" and not has_accelerate:
        raise Exception("Accelerate unavailable. Only on MacOS.")

if debug_level >= 2:
    print(f"Auto-detected {kernel_config} solver.")


def set_mkl_threads(nthreads):
    global mkl_threads, mkl
    from ctypes import byref, c_int, cdll

    mkl = cdll.LoadLibrary(mkl_path)
    # Set number of threads
    mkl_threads = nthreads
    mkl.mkl_set_num_threads(byref(c_int(nthreads)))
    if debug_level >= 5:
        print(f"MKL threads limited to {nthreads}")


if has_mkl:
    set_mkl_threads(mkl_threads)


# Compatibility layer for dictionary access to config attributes
# This is deprecated and will be removed in future


class MCEqConfigCompatibility(dict):
    """This class provides access to the attributes of the module as a
    dictionary, as it was in the previous versions of MCEq

    This method is deprecated and will be removed in future.
    """

    def __init__(self, namespace):
        self.__dict__.update(namespace)
        if debug_level > 1:
            warn_str = (
                "Config dictionary is deprecated. "
                + "Use config.variable instead of config['variable']"
            )
            warnings.warn(warn_str, FutureWarning)

    def __setitem__(self, key, value):
        key = key.lower()
        if key not in self.__dict__:
            raise Exception("Unknown config key", key)
        return super(MCEqConfigCompatibility, self).__setitem__(key, value)


config = MCEqConfigCompatibility(globals())


class FileIntegrityCheck:
    """
    A class to check a file integrity against provided checksum

    Attributes
    ----------
    filename : str
        path to the file
    checksum : str
        hex of sha256 checksum
    Methods
    -------
    succeeded():
        returns True if checksum and calculated checksum of the file are equal

    get_file_checksum():
        returns checksum of the file
    """

    import hashlib

    def __init__(self, filename, checksum=""):
        self.filename = filename
        self.checksum = checksum
        self.sha256_hash = self.hashlib.sha256()
        self.hash_is_calculated = False

    def _calculate_hash(self):
        if not self.hash_is_calculated:
            try:
                with open(self.filename, "rb") as file:
                    for byte_block in iter(lambda: file.read(4096), b""):
                        self.sha256_hash.update(byte_block)
                self.hash_is_calculated = True
            except OSError as ex:
                print(f"FileIntegrityCheck: {ex}")

    def succeeded(self):
        self._calculate_hash()
        return self.hash_is_calculated and self.sha256_hash.hexdigest() == self.checksum

    def get_file_checksum(self):
        self._calculate_hash()
        return self.sha256_hash.hexdigest()


def _download_file(url, outfile):
    """Downloads the MCEq database from github"""

    import math

    import requests
    from tqdm import tqdm

    # Streaming, so we can iterate over the response.
    r = requests.get(url, stream=True)

    # Total size in bytes.
    total_size = int(r.headers.get("content-length", 0))
    block_size = 1024 * 1024
    wrote = 0
    with open(outfile, "wb") as f:
        for data in tqdm(
            r.iter_content(block_size),
            total=math.ceil(total_size // block_size),
            unit="MB",
            unit_scale=True,
        ):
            wrote = wrote + len(data)
            f.write(data)
    if total_size != 0 and wrote != total_size:
        raise Exception("ERROR, something went wrong")


# Download database file from github
base_url = "https://github.com/afedynitch/MCEq/releases/download/"
release_tag = "builds_on_azure/"
url = base_url + release_tag + mceq_db_fname
# sha256 checksum of the file
# https://github.com/afedynitch/MCEq/releases/download/builds_on_azure/mceq_db_lext_dpm191_v12.h5
file_checksum = "5da415e9bcf81926b1061d5792d75cb3aceb9de173beccb4695fd3909a0bfdd0"

filepath_to_database = data_dir / mceq_db_fname
if filepath_to_database.exists():
    is_file_complete = FileIntegrityCheck(
        filepath_to_database, file_checksum
    ).succeeded()
else:
    is_file_complete = False

if not is_file_complete:
    print(f"Downloading for mceq database file {mceq_db_fname}.")
    if debug_level >= 2:
        print(url)
    _download_file(url, filepath_to_database)

old_database = "mceq_db_lext_dpm191.h5"
filepath_to_old_database = data_dir / old_database

if filepath_to_old_database.exists():
    import os

    print(f"Removing previous database {old_database}.")
    os.unlink(filepath_to_old_database)



import pathlib
from dataclasses import dataclass
from typing import Dict, Generator, List, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
from scipy.integrate import quad

from MCEq import config

from .ddm_utils import _eval_spline, _generate_DDM_matrix, fmteb
from .misc import info
from .particlemanager import _pdata

# isospin symmetries used in the DDM
isospin_partners = {2212: 2112, -211: 211}
isospin_symmetries = {
    2112: {
        321: 321,  # This is not exact but this is what we have
        211: -211,
        2212: 2112,
        2112: 2212,
        -2212: -2212,
        -2112: -2112,
        -211: 211,
        -321: -321,
        310: 310,
        130: 130,
    },
    211: {
        321: -321,
        211: -211,
        2212: 2112,
        2112: 2212,
        -2212: -2112,
        -2112: -2212,
        -211: 211,
        -321: 321,
        310: 310,
        130: 130,
    },
}


@dataclass
class _DDMEntry:
    """
    Represents information about a single DDM spline, i.e. for one
    specific beam energy and projectile, secondary combination.

    Attributes
    ----------
    ebeam : float
        The beam energy in GeV.
    projectile : int
        The PDG code of the projectile.
    secondary : int
        The PDG code of the secondary.
    x17 : bool
        Whether the spline fits x^1.7 dn/dx or just dn/dx.
    tck : Tuple
        The knots of the spline.
    cov : npt.NDArray
        The covariance matrix of the spline.
    tv : float
        The tuning value.
    te : float
        The error scale.
    spl_idx : int
        The index of the spline.
    """

    ebeam: str
    projectile: int
    secondary: int
    x17: bool
    tck: Tuple
    cov: npt.NDArray[np.float64]
    tv: float
    te: float
    spl_idx: int

    @property
    def knot_sigma(self) -> npt.NDArray:
        """
        Compute the knot standard deviation for the spline.

        Returns
        -------
        npt.NDArray
            The standard deviation of the knots.
        """
        return self.te * np.sqrt(np.diag(self.cov))

    @property
    def fl_ebeam(self) -> float:
        """
        Get the beam energy as a float.

        Returns
        -------
        float
            The beam energy.
        """
        return float(self.ebeam)

    @property
    def n_knots(self) -> int:
        """
        Get the number of knots in the spline.

        Returns
        -------
        int
            The number of knots.
        """
        return len(self.tck[1])

    @property
    def x_min(self) -> float:
        """
        Calculate the minimum x value for the spline.

        Returns
        -------
        float
            The minimum x value.
        """
        return _pdata.mass(self.secondary) / float(self.ebeam)

    def calc_zfactor_and_error(
        self,
        gamma: float = 1.7,
    ):
        """
        Calculate the Z-factor and its error for a given projectile,
        secondary, and beam energy.

        Parameters
        ----------
        ddm: DataDrivenModel
            Instance of DataDrivenModel.
        gamma : float, optional
            CR nucleon integral spectral index. Default is 1.7.

        Returns
        -------
        numpy.ndarray
            Z-factor values.
        numpy.ndarray
            Error of the Z-factor values.

        Notes
        -----
        The Z-factor is calculated by integrating the spline defined by the DDM entry
        over the range from entry.x_min to 1.0. The integral is multiplied by the tuning
        value of the entry and raised to the power of gamma.

        The error of the Z-factor is propagated using the covariance matrix of the
        spline and the provided _PROPAGATE_PARAMS.

        Examples
        --------
        >>> model = DataDrivenModel(...)
        >>> entry = model.spline_db.get_entry(
                projectile, secondary, ebeam=ebeam)
        >>> z_factor, z_error = entry.calc_zfactor_and_error()
        """
        from jacobi import propagate

        from MCEq import ddm_utils

        info(
            3,
            f"Calculating Z-factor for {self.projectile} --> "
            + f"{self.secondary} @ {self.ebeam} GeV.",
        )

        def func_int(tck_1):
            res = quad(
                _eval_spline,
                self.x_min,
                1.0,
                args=(
                    (self.tck[0], tck_1, self.tck[2]),
                    self.x17,
                    self.cov,
                    False,
                    gamma,
                ),
                **ddm_utils._QUAD_PARAMS,
            )[0]
            res = np.atleast_1d(res)
            res[(res < 0) | ~np.isfinite(res) | (res > ddm_utils._LIMIT_PROPAGATE)] = (
                0.0
            )
            return res.squeeze()

        y, C = propagate(func_int, self.tck[1], self.cov, **ddm_utils._PROPAGATE_PARAMS)
        sig_y = np.sqrt(C)

        return y, sig_y

    def calc_zfactor_and_error2(
        self,
        gamma: float = 1.7,
    ):
        """
        Alternative Z-factor and error calculation for a given projectile,
        secondary, and beam energy.

        Parameters
        ----------
        ddm: DataDrivenModel
            Instance of DataDrivenModel.
        gamma : float, optional
            CR nucleon integral spectral index. Default is 1.7.

        Returns
        -------
        numpy.ndarray
            Z-factor values.
        numpy.ndarray
            Error of the Z-factor values.

        Notes
        -----
        The main difference to the previous function is the error calculation, which
        is performed by integrating the error band of the evaluated splines and not
        by propagating the errors of the integral. This method is a bit more stable
        for the "problematic" splines.

        Examples
        --------
        >>> model = DataDrivenModel(...)
        >>> entry = model.spline_db.get_entry(
                projectile, secondary, ebeam=ebeam)
        >>> z_factor, z_error = entry.calc_zfactor_and_error2()
        """
        from MCEq import ddm_utils

        info(
            3,
            f"Calculating Z-factor for {self.projectile} --> "
            + f"{self.secondary} @ {self.ebeam} GeV.",
        )

        def fitfunc_center(*args, **kwargs):
            return _eval_spline(*args, **kwargs)[0]

        def fitfunc_error(*args, **kwargs):
            return _eval_spline(*args, **kwargs)[1]

        zfactor_center = quad(
            fitfunc_center,
            self.x_min,
            1.0,
            args=(self.tck, self.x17, self.cov, True, gamma),
            **ddm_utils._QUAD_PARAMS,
        )[0]
        zfactor_error = quad(
            fitfunc_error,
            self.x_min,
            1.0,
            args=(self.tck, self.x17, self.cov, True, gamma),
            **ddm_utils._QUAD_PARAMS,
        )[0]

        return (
            self.tv * zfactor_center + self.tv * self.te * zfactor_error,
            self.te * self.tv * zfactor_error,
        )


class _DDMChannel:
    """
    A class for storing DDM splines for a specific projectile, secondary.

    Attributes
    ----------
    _entries : List[_DDMEntry]
        A list of _DDMEntry objects.
    """

    def __init__(
        self,
        projectile: int,
        secondary: int,
    ):
        self._entries: List[_DDMEntry] = []
        self.projectile: int = projectile
        self.secondary: int = secondary

    def add_entry(
        self,
        ebeam: Union[float, str],
        projectile: int,
        secondary: int,
        x17: bool,
        tck: Tuple,
        cov: npt.NDArray,
        tv: float = 1.0,
        te: float = 1.0,
    ):
        """
        Adds a new DDM entry.

        Parameters
        ----------
        ebeam : float
            The beam energy in GeV.
        projectile : int
            The PDG code of the projectile.
        secondary : int
            The PDG code of the secondary.
        x17 : bool
            Whether the spline fits x^1.7 dn/dx or just dn/dx.
        tck : Tuple
            The knots of the spline.
        cov : npt.NDArray
            The covariance matrix of the spline.
        tv : float, optional
            The tuning value, by default 1.0.
        te : float, optional
            The error scale, by default 1.0.
        """

        assert projectile == self.projectile
        assert secondary == self.secondary

        for entry in self._entries:
            if fmteb(entry.ebeam) == fmteb(ebeam):
                raise ValueError(f"Duplicate entry for ebeam = {ebeam} GeV.")
        info(10, f"Adding DDM spline for {projectile} -> {secondary} at {ebeam} GeV")
        self._entries.append(
            _DDMEntry(
                fmteb(ebeam),
                projectile,
                secondary,
                x17,
                tck,
                cov,
                tv,
                te,
                len(self._entries),
            )
        )

        self._entries.sort(key=lambda entry: float(entry.ebeam))

        # Reassign the spline indices
        for i, entry in enumerate(self._entries):
            entry.spl_idx = i

    @property
    def entries(self) -> List[_DDMEntry]:
        """
        Returns the DDM entries.

        Returns
        -------
        List[_DDMEntry]
            A list of _DDMEntry objects.
        """

        return self._entries

    @property
    def total_n_knots(self) -> int:
        """
        Returns the total number of knots.

        Returns
        -------
        int
            The total number of knots.
        """

        return sum([len(entry.tck[0]) for entry in self._entries])

    @property
    def n_splines(self) -> int:
        """
        Returns the number of splines.

        Returns
        -------
        int
            The number of splines.
        """

        return len(self._entries)

    @property
    def spline_indices(self) -> List[int]:
        """
        Returns the spline indices.

        Returns
        -------
        List[int]
            A list of spline indices.
        """
        return [entry.spl_idx for entry in self._entries]

    def get_entry(
        self,
        ebeam: Optional[Union[float, int, str]] = None,
        idx: Optional[int] = None,
    ) -> _DDMEntry:
        """
        Returns a DDM entry.

        Parameters
        ----------
        ebeam : Optional[Union[float, int, str]], optional
            The beam energy in GeV, by default None.
        idx : Optional[int], optional
            The index of the spline, by default None.

        Returns
        -------
        _DDMEntry
            The DDM entry.

        Raises
        ------
        ValueError
            If no entry is found for the given ebeam or spline index.
        """

        assert (ebeam is not None) != (idx is not None), (
            "Define either ebeam or spl_idx"
        )

        if ebeam is not None:
            ebeam = fmteb(ebeam)
            for entry in self._entries:
                if entry.ebeam == ebeam:
                    return entry
            raise ValueError(f"No entry for ebeam = {ebeam} GeV.")
        else:
            for entry in self._entries:
                if entry.spl_idx == idx:
                    return entry
            raise ValueError(f"No entry for spl_idx = {idx}.")

    def __str__(self) -> str:
        s = f"\t{self.projectile} -> {self.secondary}:\n"
        for iset, dc in enumerate(self.entries):
            s += f"\t\t{iset}: ebeam = {dc.ebeam} GeV, x17={dc.x17}, "
            s += f"tune v|e={dc.tv:4.3f}|{dc.te:4.3f}\n"
        return s


class DDMSplineDB:
    """A class for maintaining DDMEntryCollections for different
    projectile and secondary particle combinations."""

    _ddm_splines: Dict[str, _DDMChannel] = {}

    def __init__(
        self,
        filename: str = str(pathlib.Path(config.data_dir) / "DDM_1.0.npy"),
        enable_channels: List[Tuple[int, int]] = [],
        exclude_projectiles: List[int] = [],
    ):
        """
        Initializes the DDMSplineDB.

        Parameters
        ----------
        filename : str, optional
            The filename of the DDM spline data file, by default "DDM_1.0.npy".
        enable_channels : List[Tuple[int, int]], optional
            List of projectile and secondary combinations to enable,
            by default an empty list.
        exclude_projectiles : List[int], optional
            List of projectile codes to exclude, by default an empty list.
        """
        self._load_from_file(filename, enable_channels, exclude_projectiles)

    def _load_from_file(
        self,
        filename: str,
        enable_channels: List[Tuple[int, int]],
        exclude_projectiles: List[int],
    ):
        """
        Loads DDM splines from a file.

        Parameters
        ----------
        filename : str
            The filename of the DDM spline data file.
        enable_channels : List[Tuple[int, int]]
            List of projectile and secondary combinations to enable.
        exclude_projectiles : List[int]
            List of projectile codes to exclude.
        """

        # Clean up if _load is called multiple times
        if self._ddm_splines:
            self._ddm_splines = {}

        spl_file = np.load(filename, allow_pickle=True, encoding="latin1").item()
        for projectile, secondary, ebeam, x17 in spl_file:
            if (enable_channels and (projectile, secondary) not in enable_channels) or (
                abs(projectile) in exclude_projectiles
            ):
                continue
            else:
                self.add_entry(
                    ebeam,
                    projectile,
                    secondary,
                    x17,
                    spl_file[(projectile, secondary, ebeam, x17)][0],
                    spl_file[(projectile, secondary, ebeam, x17)][1],
                    1.0,
                    1.0,
                )

    def add_entry(
        self,
        ebeam: Union[float, str],
        projectile: int,
        secondary: int,
        x17: bool,
        tck: Tuple,
        cov: npt.NDArray,
        tv: float = 1.0,
        te: float = 1.0,
    ):
        """
        Adds a new DDM spline.

        Parameters
        ----------
        ebeam : Union[float, str]
            The beam energy in GeV.
        projectile : int
            The projectile.
        secondary : int
            The secondary particle.
        x17 : bool
            Whether the spline fits x^1.7 dn/dx or just dn/dx.
        tck : Tuple
            The knots of the spline.
        cov : npt.NDArray
            The covariance matrix of the spline.
        tv : float, optional
            The tuning value, by default 1.0.
        te : float, optional
            The error scale, by default 1.0.
        """
        info(5, f"Adding DDM spline for {projectile} -> {secondary} at {ebeam} GeV")

        channel = self._mk_channel(projectile, secondary)
        if channel not in self._ddm_splines:
            self._ddm_splines[channel] = _DDMChannel(projectile, secondary)

        self._ddm_splines[channel].add_entry(
            fmteb(ebeam), projectile, secondary, x17, tck, cov, tv, te
        )

    def get_spline_indices(self, projectile: int, secondary: int) -> List[int]:
        """
        Returns the spline indices.

        Parameters
        ----------
        projectile : int
            The projectile.
        secondary : int
            The secondary particle.
        ebeam : Union[float, str, int]
            The beam energy in GeV.

        Returns
        -------
        List[int]
            A list of spline indices.
        """
        channel = self._mk_channel(projectile, secondary)
        return [ent.spl_idx for ent in self._ddm_splines[channel].entries]

    @property
    def channels(self) -> Generator[_DDMChannel, None, None]:
        """
        Returns the DDM channels.

        Yields
        ------
        _DDMChannel
            A generator of _DDMChannel objects.
        """
        for channel in self._ddm_splines:
            yield self._ddm_splines[channel]

    def clone_entry(
        self, projectile: int, secondary: int, original_ebeam: float, new_ebeam: float
    ) -> None:
        """
        Clones a DDM entry.

        Parameters
        ----------
        projectile : int
            The projectile.
        secondary : int
            The secondary particle.
        original_ebeam : float
            The original beam energy in GeV.
        new_ebeam : float
            The new beam energy in GeV.
        """
        entry = self.get_entry(projectile, secondary, original_ebeam)
        self.add_entry(
            fmteb(new_ebeam),
            entry.projectile,
            entry.secondary,
            entry.x17,
            entry.tck,
            entry.cov,
            1.0,  # Don't copy tuning values to cloned entry
            1.0,
        )

    def get_entry(
        self,
        projectile: int,
        secondary: int,
        ebeam: Optional[float] = None,
        idx: Optional[int] = None,
    ) -> _DDMEntry:
        """
        Returns a DDM entry.

        Parameters
        ----------
        projectile : int
            The projectile.
        secondary : int
            The secondary particle.
        ebeam : Optional[float], optional
            The beam energy in GeV, by default None.
        idx : Optional[int], optional
            The index of the spline, by default None.

        Returns
        -------
        _DDMEntry
            The DDM entry.
        """
        return self._ddm_splines[self._mk_channel(projectile, secondary)].get_entry(
            ebeam, idx
        )

    def _mk_channel(self, projectile: int, secondary: int) -> str:
        """
        Creates a channel key.

        Parameters
        ----------
        projectile : int
            The projectile.
        secondary : int
            The secondary particle.

        Returns
        -------
        str
            The channel key.
        """
        return f"{projectile}-{secondary}"


class DataDrivenModel:
    """This class implements the Data-Driven Model."""

    def __init__(
        self,
        filename: str = str(pathlib.Path(config.data_dir) / "DDM_1.0.npy"),
        e_min: float = -1.0,
        e_max: float = -1.0,
        enable_channels: List[Tuple[int, int]] = [],
        exclude_projectiles: List[int] = [],
        enable_K0_from_isospin: bool = True,
    ):
        """
        Initializes a DataDrivenModel object.

        Parameters
        ----------
        filename : str, optional
            The filename of the data file used to build the model,
            by default "DDM_1.0.npy".
        e_min : float, optional
            The minimum energy range where DDM cross sections
            overwrite original MCEq matrices, by default 0.
        e_max : float, optional
            The maximum energy range where DDM cross sections
            overwrite original MCEq matrices, by default np.inf.
        enable_channels : List[Tuple[int, int]], optional
            List of projectile and secondary combinations to enable,
            by default an empty list.
        exclude_projectiles : List[int], optional
            List of projectile codes to exclude, by default an empty list.
        enable_K0_from_isospin : bool, optional
            Whether to use isospin relation for K0S/L, by default True.
        """
        self.e_min = e_min
        self.e_max = e_max
        self.enable_K0_from_isospin = enable_K0_from_isospin
        self.spline_db = DDMSplineDB(filename, enable_channels, exclude_projectiles)

    def ddm_matrices(self, mceq) -> Dict[Tuple[int, int], npt.NDArray]:
        """
        Generates a dictionary of DDM matrices for the given MCEq object.

        Parameters
        ----------
        mceq : object
            An MCEq object.

        Returns
        -------
        Dict[Tuple[int, int], npt.NDArray]
            A dictionary with keys as tuples (projectile PDG, secondary PDG) and
            values as numpy arrays representing the corresponding DDM matrices.
        """

        _ddm_mat = dict()
        for channel in self.spline_db.channels:
            info(
                0, f"Generating {channel.projectile} -> {channel.secondary} DDM matrix"
            )
            _ddm_mat[(channel.projectile, channel.secondary)] = _generate_DDM_matrix(
                channel=channel, mceq=mceq, e_min=mceq.e_bins[0], e_max=mceq.e_bins[-1]
            )

        if self.enable_K0_from_isospin:
            info(3, "Generating DDM K0 matrices from isospin symm.")
            assert (2212, 321) in _ddm_mat, f"Missing K+ matrices in {_ddm_mat.keys()}"
            K0SL_mat = 0.5 * (_ddm_mat[(2212, 321)] + _ddm_mat[(2212, -321)])
            _ddm_mat[(2212, 310)] = K0SL_mat
            _ddm_mat[(2212, 130)] = K0SL_mat
            if (-211, 321) in _ddm_mat and (-211, -321) in _ddm_mat:
                K0SL_mat = 0.5 * (_ddm_mat[(-211, 321)] + _ddm_mat[(-211, -321)])
                _ddm_mat[(-211, 310)] = K0SL_mat
                _ddm_mat[(-211, 130)] = K0SL_mat

        return _ddm_mat

    def apply_tuning(
        self,
        projectile: int,
        secondary: int,
        ebeam: Optional[float] = None,
        spl_idx: Optional[int] = None,
        tv: float = 1.0,
        te: float = 1.0,
    ) -> None:
        """
        Modify the tuning values and trim errors in the data_combinations dictionary.

        Parameters
        ----------
        projectile : int
            Primary particle code.
        secondary : int
            Secondary particle code.
        ebeam : Optional[float], optional
            Lab frame beam energy in GeV, by default None.
        spl_idx : Optional[int], optional
            Index of the spline to modify, by default None.
        tv : float, optional
            Scaling factor for the central value, by default 1.0.
        te : float, optional
            Scaling factor for the errors, by default 1.0.

        Raises
        ------
        AssertionError
            If neither `spl_idx` nor `ebeam` is set.
        """

        entry = self.spline_db.get_entry(projectile, secondary, ebeam, spl_idx)
        entry.tv = tv
        entry.te = te

    def dn_dxl(
        self,
        x: np.ndarray,
        projectile: int,
        secondary: int,
        ebeam: float,
        return_error: bool = True,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """
        Returns dN/dxL and error.

        Parameters
        ----------
        x : np.ndarray
            Array of x values.
        projectile : int
            The projectile.
        secondary : int
            The secondary particle.
        ebeam : float
            Lab frame beam energy in GeV.
        return_error : bool, optional
            Whether to return the error, by default True.

        Returns
        -------
        Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]
            The dN/dxL values and optional error.

        """
        entry = self.spline_db.get_entry(projectile, secondary, ebeam)

        res = _eval_spline(
            x, entry.tck, entry.x17, entry.cov, return_error=return_error
        )
        mask = np.where(x < _pdata.mass(secondary) / ebeam)
        if isinstance(res, tuple):
            v, e = res
            v[mask] = 0.0
            e[mask] = 0.0
            return v, e
        elif isinstance(res, np.ndarray):
            res[mask] = 0

        return res

    def find_channel(self, projectile: int, secondary: int) -> _DDMChannel:
        """
        Finds a DDM channel.

        Parameters
        ----------
        projectile : int
            The projectile.
        secondary : int
            The secondary particle.

        Returns
        -------
        _DDMChannel
            The DDM channel.
        """
        return self.spline_db._ddm_splines[
            self.spline_db._mk_channel(projectile, secondary)
        ]

    def __repr__(self) -> str:
        s = "DDM channels:\n"
        for channel in self.spline_db.channels:
            s += str(channel)
        s += "\n"
        return s



from math import copysign

import numpy as np
import six
from particletools.tables import PYTHIAParticleData

from MCEq import config
from MCEq.misc import average_A_target, getAZN, info, print_in_rows

info(5, "Initialization of PYTHIAParticleData object")
_pdata = PYTHIAParticleData()

backward_compatible_namestr = {
    "nu_mu": "numu",
    "nu_mubar": "antinumu",
    "nu_e": "nue",
    "nu_ebar": "antinue",
    "nu_tau": "nutau",
    "nu_taubar": "antinutau",
}


# Replace particle names for neutrinos with those used
# in previous MCEq versions
def _pname(pdg_id_or_name):
    """Replace some particle names from pythia database with those from previous
    MCEq versions for backward compatibility."""

    pythia_name = _pdata.name(pdg_id_or_name)
    if pythia_name in backward_compatible_namestr:
        return backward_compatible_namestr[pythia_name]
    return pythia_name


class MCEqParticle:
    """Bundles different particle properties for simplified
    availability of particle properties in :class:`MCEq.core.MCEqRun`.

    Args:
      pdg_id (int): PDG ID of the particle
      egrid (np.array, optional): energy grid (centers)
      cs_db (object, optional): reference to an instance of
                                :class:`InteractionYields`
    """

    def __init__(
        self,
        pdg_id,
        helicity,
        energy_grid=None,
        cs_db=None,
        init_pdata_defaults=True,
        A_target=average_A_target(),
    ):
        #: (bool) if it's an electromagnetic particle
        self.is_em = abs(pdg_id) == 11 or pdg_id == 22
        #: (int) helicity -1, 0, 1 (0 means undefined or average)
        self.helicity = helicity
        #: (bool) particle is a nucleus (not yet implemented)
        self.is_nucleus = False
        #: (bool) particle is a hadron
        self.is_hadron = False
        #: (bool) particle is a lepton
        self.is_lepton = False
        #: (bool) particle is charged
        self.is_charged = False
        #: (float) ctau in cm
        self.ctau = None
        #: (float) mass in GeV
        self.mass = None
        #: (int) Charge
        self.charge = None
        #: (str) species name in string representation
        self.name = None
        #: Mass, charge, neutron number
        self.A, self.Z, self.N = getAZN(pdg_id)
        #: (bool) particle has both, hadron and resonance properties
        self.is_mixed = False
        #: (bool) if particle has just resonance behavior
        self.is_resonance = False
        #: (bool) particle is interacting projectile
        self.is_projectile = False
        #: (bool) particle is stable
        self.is_stable = False or pdg_id in config.adv_set["disable_decays"]
        #: (bool) can_interact
        self.can_interact = False
        #: (bool) has continuous losses dE/dX defined
        self.has_contloss = False
        #: (np.array) continuous losses in GeV/(g/cm2)
        self.dEdX = None
        #: (bool) is a tracking particle
        self.is_tracking = False
        #: decay channels if any
        self.decay_dists = {}
        #: (int) Particle Data Group Monte Carlo particle ID
        self.pdg_id = (pdg_id, helicity)
        #: (int) Unique PDG ID that is different for tracking particles
        self.unique_pdg_id = (pdg_id, helicity)
        #: (int) MCEq ID
        self.mceqidx = -1

        #: (float) mixing energy, transition between hadron and
        # resonance behavior
        self.E_mix = 0
        #: (int) energy grid index, where transition between
        # hadron and resonance occurs
        self.mix_idx = 0
        #: (float) interaction threshold (idx where cs>0)
        self.int_idx = 0
        #: (float) critical energy in air at the surface
        self.E_crit = 0

        # Energy and cross section dependent inits
        self.current_cross_sections = None
        self._energy_grid = energy_grid

        # Variables for hadronic interaction
        self.current_hadronic_model = None
        self.hadr_secondaries = []
        self.hadr_yields = {}

        # Variables for decays
        self.children = []
        self.decay_dists = {}

        # A_target
        self.A_target = A_target

        if init_pdata_defaults:
            self._init_defaults_from_pythia_database()

        if self._energy_grid is not None and cs_db is not None:
            #: interaction cross section in 1/cm2
            self.set_cs(cs_db)

    def _init_defaults_from_pythia_database(self):
        """Init some particle properties from :mod:`particletools.tables`."""
        #: (bool) particle is a nucleus (not yet implemented)
        self.is_nucleus = _pdata.is_nucleus(self.pdg_id[0])
        #: (bool) particle is a hadron
        self.is_hadron = _pdata.is_hadron(self.pdg_id[0])
        #: (bool) particle is a hadron
        self.is_lepton = _pdata.is_lepton(self.pdg_id[0])
        #: Mass, charge, neutron number
        self.A, self.Z, self.N = getAZN(self.pdg_id[0])
        if self.is_nucleus:
            self.charge = self.Z
        else:
            self.charge = int(_pdata.charge(self.pdg_id[0]))
        self.is_charged = bool(abs(self.charge))
        #: (float) ctau in cm
        self.ctau = _pdata.ctau(self.pdg_id[0])
        #: (float) mass in GeV
        self.mass = _pdata.mass(self.pdg_id[0])
        #: (str) species name in string representation
        name = _pname(self.pdg_id[0]) if self.name is None else self.name
        if self.helicity == -1:
            name += "_l"
        elif self.helicity == +1:
            name += "_r"
        self.name = name
        #: (bool) particle is stable
        #: TODO the exclusion of neutron decays is a hotfix
        self.is_stable = (
            not self.ctau < np.inf or self.pdg_id[0] in config.adv_set["disable_decays"]
        )

    def init_custom_particle_data(self, name, pdg_id, helicity, ctau, mass, **kwargs):
        """Add custom particle type. (Incomplete and not debugged)"""
        #: (int) Particle Data Group Monte Carlo particle ID
        self.pdg_id = (pdg_id, helicity)
        #: (bool) if it's an electromagnetic particle
        self.is_em = kwargs.pop("is_em", abs(pdg_id) == 11 or pdg_id == 22)
        #: (bool) particle is a nucleus (not yet implemented)
        self.is_nucleus = kwargs.pop("is_nucleus", _pdata.is_nucleus(self.pdg_id[0]))
        #: (bool) particle is a hadron
        self.is_hadron = kwargs.pop("is_hadron", _pdata.is_hadron(self.pdg_id[0]))
        #: (bool) particle is a hadron
        self.is_lepton = kwargs.pop("is_lepton", _pdata.is_lepton(self.pdg_id[0]))
        #: Mass, charge, neutron number
        self.A, self.Z, self.N = getAZN(self.pdg_id[0])
        #: (float) ctau in cm
        self.ctau = ctau
        #: (float) mass in GeV
        self.mass = mass
        #: (str) species name in string representation
        self.name = name
        #: (bool) particle is stable
        self.is_stable = not self.ctau < np.inf

    def set_cs(self, cs_db):
        """Set cross section adn recalculate the dependent variables"""
        info(11, "Obtaining cross sections for", self.pdg_id)
        self.current_cross_sections = cs_db.iam
        self.cs = cs_db[self.pdg_id[0]]
        if sum(self.cs) > 0:
            self.can_interact = True
            # self._interaction_threshold()
        else:
            self.can_interact = False

        self._critical_energy()
        self._calculate_mixing_energy()

    def set_hadronic_channels(self, hadronic_db, pmanager):
        """Changes the hadronic interaction model.

        Replaces indexing of the yield dictionary from PDG IDs
        with references from partilcle manager.
        """

        self.current_hadronic_model = hadronic_db.iam
        # Collect MCEqParticle references to children
        # instead of PDG ID as index
        # Also copy over tracking relations if they exist
        tracking_relations = [
            tr_ref for tr_ref in self.hadr_secondaries if tr_ref.is_tracking
        ]
        if self.pdg_id in hadronic_db.parents and not self.is_tracking:
            self.is_projectile = True
            self.hadr_secondaries = [
                pmanager.pdg2pref[pid] for pid in hadronic_db.relations[self.pdg_id]
            ] + tracking_relations
            self.hadr_yields = {}
            for s in self.hadr_secondaries:
                self.hadr_yields[s] = hadronic_db.get_matrix(self.pdg_id, s.pdg_id)
        else:
            self.is_projectile = False
            self.hadr_secondaries = []
            self.hadr_yields = {}

        if self.is_tracking:
            # Copy properties of the original particle for tracking
            orig_part = pmanager.pdg2pref[self.pdg_id]
            self.is_projectile = orig_part.is_projectile
            if orig_part in orig_part.hadr_secondaries:
                self.hadr_secondaries.append(self)
                self.hadr_yields[self] = orig_part.hadr_yields[orig_part]

    def add_hadronic_production_channel(self, child, int_matrix):
        """Add a new particle that is produced in hadronic interactions.

        The int_matrix is expected to be in the correct shape and scale
        as the other interaction (dN/dE(i,j)) matrices. Energy conservation
        is not checked.
        """

        if not self.is_projectile:
            raise Exception("The particle should be a projectile.")

        if child in self.hadr_secondaries:
            info(1, f"Child {child.name} has been already added.")
            return

        self.hadr_secondaries.append(child)
        self.hadr_yields[child] = int_matrix

    def add_decay_channel(self, child, dec_matrix, force=False):
        """Add a decay channel.

        The branching ratios are not renormalized and one needs to take care
        of this externally.
        """
        if self.is_stable:
            raise Exception("Cannot add decay channel to stable particle.")

        if child in self.children and not force:
            info(1, f"Child {child.name} has been already added.")
            return
        if child in self.children and force:
            info(1, f"Overwriting decay matrix of child {child.name}.")
            self.decay_dists[child] = dec_matrix
            return

        self.children.append(child)
        self.decay_dists[child] = dec_matrix

    def set_decay_channels(self, decay_db, pmanager):
        """Populates decay channel and energy distributions"""

        if self.is_stable or self.is_tracking:
            # Variables for decays
            self.children = []
            self.decay_dists = {}
            return

        if self.pdg_id not in decay_db.parents:
            raise Exception(
                "Unstable particle without decay distribution:", self.pdg_id, self.name
            )

        self.children = []
        self.children = [pmanager[d] for d in decay_db.children(self.pdg_id)]
        self.decay_dists = {}
        for c in self.children:
            self.decay_dists[c] = decay_db.get_matrix(self.pdg_id, c.pdg_id)

    def track_decays(self, tracking_particle):
        children_d = {}
        for c in self.children:
            children_d[c.pdg_id] = c
        if tracking_particle.pdg_id not in list(children_d):
            _n = self.name
            _tn = tracking_particle.name
            info(
                17,
                f"Parent particle {_n} does not decay into {_tn}",
            )
            return False
        # Copy the decay distribution from original PDG
        self.children.append(tracking_particle)
        self.decay_dists[tracking_particle] = self.decay_dists[
            children_d[tracking_particle.pdg_id]
        ]
        return True

    def track_interactions(self, tracking_particle):
        secondaries_d = {}
        for s in self.hadr_secondaries:
            secondaries_d[s.pdg_id] = s
        if tracking_particle.pdg_id not in list(secondaries_d):
            _n = self.name
            _tn = tracking_particle.name
            info(
                17,
                f"Parent particle {_n} does not produce {_tn} at the vertex",
            )
            return False
        # Copy the interaction matrix from original PDG
        self.hadr_secondaries.append(tracking_particle)
        self.hadr_yields[tracking_particle] = self.hadr_yields[
            secondaries_d[tracking_particle.pdg_id]
        ]
        return True

    def is_secondary(self, particle_ref):
        """`True` if this projectile and produces particle `particle_ref`."""
        if not isinstance(particle_ref, self.__class__):
            raise Exception("Argument not of MCEqParticle type.")
        return particle_ref in self.hadr_secondaries

    def is_child(self, particle_ref):
        """`True` if this particle decays into `particle_ref`."""
        if not isinstance(particle_ref, self.__class__):
            raise Exception("Argument not of MCEqParticle type.")
        return particle_ref in self.children

    @property
    def hadridx(self):
        """Returns index range where particle behaves as hadron.

        Returns:
          :func:`tuple` (int,int): range on energy grid
        """
        return (self.mix_idx, self._energy_grid.d)

    @property
    def residx(self):
        """Returns index range where particle behaves as resonance.

        Returns:
          :func:`tuple` (int,int): range on energy grid
        """
        return (0, self.mix_idx)

    @property
    def lidx(self):
        """Returns lower index of particle range in state vector.

        Returns:
          (int): lower index in state vector :attr:`MCEqRun.phi`
        """
        return self.mceqidx * self._energy_grid.d

    @property
    def uidx(self):
        """Returns upper index of particle range in state vector.

        Returns:
          (int): upper index in state vector :attr:`MCEqRun.phi`
        """
        return (self.mceqidx + 1) * self._energy_grid.d

    def inverse_decay_length(self, cut=True):
        r"""Returns inverse decay length (or infinity (np.inf), if
        particle is stable), where the air density :math:`\rho` is
        factorized out.

        Args:
          E (float) : energy in laboratory system in GeV
          cut (bool): set to zero in 'resonance' regime
        Returns:
          (float): :math:`\frac{\rho}{\lambda_{dec}}` in 1/cm
        """
        try:
            dlen = self.mass / self.ctau / (self._energy_grid.c + self.mass)
            if cut:
                dlen[0 : self.mix_idx] = 0.0
            # Correction for bin average, since dec. length is a steep falling
            # function. This factor averages the value over bin length for
            # 10 bins per decade.
            # return 0.989 * dlen
            return dlen
        except ZeroDivisionError:
            return np.ones_like(self._energy_grid.d) * np.inf

    def inel_cross_section(self, mbarn=False):
        """Returns inelastic cross section.

        Args:
          mbarn (bool) : if True cross section in mb otherwise in cm**2
        Returns:
          (float): :math:`\\sigma_{\\rm inel}` in mb or cm**2
        """
        #: unit - :math:`\text{GeV} \cdot \text{fm}`
        GeVfm = 0.19732696312541853
        #: unit - :math:`\text{GeV} \cdot \text{cm}`
        GeVcm = GeVfm * 1e-13
        #: unit - :math:`\text{GeV}^2 \cdot \text{mbarn}`
        GeV2mbarn = 10.0 * GeVfm**2
        #: unit conversion - :math:`\text{mbarn} \to \text{cm}^2`
        mbarn2cm2 = GeV2mbarn / GeVcm**2
        if mbarn:
            return mbarn2cm2 * self.cs

        return self.cs

    def inverse_interaction_length(self):
        """Returns inverse interaction length for A_target given by config.

        Returns:
          (float): :math:`\\frac{1}{\\lambda_{int}}` in cm**2/g
        """

        m_target = self.A_target * 1.672621 * 1e-24  # <A> * m_proton [g]
        return self.cs / m_target

    def _assign_hadr_dist_idx(self, child, projidx, chidx, cmat):
        """Copies a subset, defined between indices ``projidx`` and ``chiidx``
        from the ``hadr_yields`` into ``cmat``

        Args:
          child (int): PDG ID of final state child/secondary particle
          projidx (int,int): tuple containing index range relative
                             to the projectile's energy grid
          dtridx (int,int): tuple containing index range relative
                            to the child's energy grid
          cmat (:func:`numpy.array`): array reference to the interaction matrix
        """
        cmat[chidx[0] : chidx[1], projidx[0] : projidx[1]] = self.hadr_yields[child][
            chidx[0] : chidx[1], projidx[0] : projidx[1]
        ]

    def _assign_decay_idx(self, child, projidx, chidx, cmat):
        """Copies a subset, defined between indices ``projidx`` and ``chiidx``
        from the ``hadr_yields`` into ``cmat``

        Args:
          child (int): PDG ID of final state child/secondary particle
          projidx (int,int): tuple containing index range relative
                             to the projectile's energy grid
          dtridx (int,int): tuple containing index range relative
                            to the child's energy grid
          cmat (:func:`numpy.array`): array reference to the interaction matrix
        """
        cmat[chidx[0] : chidx[1], projidx[0] : projidx[1]] = self.decay_dists[child][
            chidx[0] : chidx[1], projidx[0] : projidx[1]
        ]

    def dN_dxlab(self, kin_energy, sec_pdg, verbose=True, **kwargs):
        r"""Returns :math:`dN/dx_{\rm Lab}` for interaction energy close
        to ``kin_energy`` for hadron-air collisions.

        The function respects modifications applied via :func:`_set_mod_pprod`.

        Args:
            kin_energy (float): approximate interaction kin_energy
            prim_pdg (int): PDG ID of projectile
            sec_pdg (int): PDG ID of secondary particle
            verbose (bool): print out the closest enerkin_energygy
        Returns:
            (:func:`numpy.array`, :func:`numpy.array`): :math:`x_{\rm Lab}`, :math:`dN/dx_{\rm Lab}`
        """

        eidx = (np.abs(self._energy_grid.c - kin_energy)).argmin()
        en = self._energy_grid.c[eidx]
        info(10, "Nearest energy, index: ", en, eidx, condition=verbose)

        m = self.hadr_yields[sec_pdg]
        xl_grid = (self._energy_grid.c[: eidx + 1]) / en

        if config.kernel_config == "CUDA":
            import cupy

            xl_dist = (
                en
                * xl_grid
                * cupy.asnumpy(m[: eidx + 1, eidx])
                / self._energy_grid.w[: eidx + 1]
            )
        else:
            xl_dist = (
                en * xl_grid * m[: eidx + 1, eidx] / self._energy_grid.w[: eidx + 1]
            )

        return xl_grid, xl_dist

    def dNdec_dxlab(self, kin_energy, sec_pdg, verbose=True, **kwargs):
        r"""Returns :math:`dN/dx_{\rm Lab}` for interaction energy close
        to ``kin_energy`` for hadron-air collisions.

        The function respects modifications applied via :func:`_set_mod_pprod`.

        Args:
            kin_energy (float): approximate interaction energy
            prim_pdg (int): PDG ID of projectile
            sec_pdg (int): PDG ID of secondary particle
            verbose (bool): print out the closest energy
        Returns:
            (:func:`numpy.array`, :func:`numpy.array`): :math:`x_{\rm Lab}`, :math:`dN/dx_{\rm Lab}`
        """

        eidx = (np.abs(self._energy_grid.c - kin_energy)).argmin()
        en = self._energy_grid.c[eidx]
        info(10, "Nearest energy, index: ", en, eidx, condition=verbose)

        m = self.decay_dists[sec_pdg]
        xl_grid = (self._energy_grid.c[: eidx + 1]) / en

        if config.kernel_config == "CUDA":
            import cupy

            xl_dist = (
                en
                * xl_grid
                * cupy.asnumpy(m[: eidx + 1, eidx])
                / self._energy_grid.w[: eidx + 1]
            )
        else:
            xl_dist = (
                en * xl_grid * m[: eidx + 1, eidx] / self._energy_grid.w[: eidx + 1]
            )

        return xl_grid, xl_dist

    def dN_dEkin(self, kin_energy, sec_pdg, verbose=True, **kwargs):
        r"""Returns :math:`dN/dE_{\rm Kin}` in lab frame for an interaction energy
        close to ``kin_energy`` (total) for hadron-air collisions.

        The function respects modifications applied via :func:`_set_mod_pprod`.

        Args:
            kin_energy (float): approximate interaction energy
            prim_pdg (int): PDG ID of projectile
            sec_pdg (int): PDG ID of secondary particle
            verbose (bool): print out the closest energy
        Returns:
            (:func:`numpy.array`, :func:`numpy.array`): :math:`x_{\rm Lab}`, :math:`dN/dx_{\rm Lab}`
        """

        eidx = (np.abs(self._energy_grid.c - kin_energy)).argmin()
        en = self._energy_grid.c[eidx]
        info(10, "Nearest energy, index: ", en, eidx, condition=verbose)

        m = self.hadr_yields[sec_pdg]
        ekin_grid = self._energy_grid.c
        if config.kernel_config == "CUDA":
            import cupy

            elab_dist = cupy.asnumpy(m[: eidx + 1, eidx]) / self._energy_grid.w[eidx]
        else:
            elab_dist = m[: eidx + 1, eidx] / self._energy_grid.w[eidx]
        return ekin_grid[: eidx + 1], elab_dist

    def dN_dxf(self, energy, prim_pdg, sec_pdg, pos_only=True, verbose=True, **kwargs):
        r"""Returns :math:`dN/dx_{\rm F}` in c.m. for interaction energy close
        to ``energy`` (lab. not kinetic) for hadron-air collisions.

        The function respects modifications applied via :func:`_set_mod_pprod`.

        Args:
            energy (float): approximate interaction lab. energy
            prim_pdg (int): PDG ID of projectile
            sec_pdg (int): PDG ID of secondary particle
            verbose (bool): print out the closest energy

        Returns:
            (:func:`numpy.array`, :func:`numpy.array`): :math:`x_{\rm F}`, :math:`dN/dx_{\rm F}`
        """
        if not hasattr(self, "_ptav_sib23c"):
            # Load spline of average pt distribution as a funtion of log(E_lab) from sib23c
            import pickle
            from os.path import join

            self._ptav_sib23c = pickle.load(
                open(join(config.data_dir, "sibyll23c_aux.ppd"), "rb")
            )[0]

        def xF(xL, Elab, ppdg):
            m = {2212: 0.938, 211: 0.139, 321: 0.493}
            mp = m[2212]

            Ecm = np.sqrt(2 * Elab * mp + 2 * mp**2)
            Esec = xL * Elab
            betacm = np.sqrt((Elab - mp) / (Elab + mp))
            gammacm = (Elab + mp) / Ecm
            avpt = self._ptav_sib23c[ppdg](
                np.log(np.sqrt(Elab**2) - m[np.abs(ppdg)] ** 2)
            )

            xf = (
                2
                * (
                    -betacm * gammacm * Esec
                    + gammacm * np.sqrt(Esec**2 - m[np.abs(ppdg)] ** 2 - avpt**2)
                )
                / Ecm
            )
            dxl_dxf = 1.0 / (
                2
                * (
                    -betacm * gammacm * Elab
                    + xL
                    * Elab**2
                    * gammacm
                    / np.sqrt((xL * Elab) ** 2 - m[np.abs(ppdg)] ** 2 - avpt**2)
                )
                / Ecm
            )

            return xf, dxl_dxf

        eidx = (np.abs(self._energy_grid.c + self.mass - energy)).argmin()
        en = self._energy_grid.c[eidx] + self.mass
        info(2, "Nearest energy, index: ", en, eidx, condition=verbose)
        m = self.hadr_yields[sec_pdg]
        xl_grid = (self._energy_grid.c[: eidx + 1] + self.mass) / en
        xl_dist = (
            xl_grid
            * en
            * m[: eidx + 1, eidx]
            / np.diag(self._energy_grid.w)[: eidx + 1]
        )
        xf_grid, dxl_dxf = xF(xl_grid, en, sec_pdg)
        xf_dist = xl_dist * dxl_dxf

        if pos_only:
            xf_dist = xf_dist[xf_grid >= 0]
            xf_grid = xf_grid[xf_grid >= 0]
            return xf_grid, xf_dist

        return xf_grid, xf_dist

    def _critical_energy(self):
        """Returns critical energy where decay and interaction
        are balanced.

        Approximate value in Air.
        """
        if self.is_stable or self.ctau <= 0.0:
            self.E_crit = np.inf
        else:
            self.E_crit = self.mass * 6.4e5 / self.ctau

    def _interaction_threshold(self):
        """Finds minimal energy grid idx where interaction cross sections
        becomes > 0.

        Only for interacting particles.
        """
        if self.can_interact:
            self.int_idx = (self.cs != 0).argmax()
            info(
                10,
                "Interaction threshold for {0} is {1:5.3f} GeV".format(
                    self.name, self._energy_grid.c[self.int_idx]
                ),
            )
        else:
            self.int_idx = 0

    def _calculate_mixing_energy(self):
        """Calculates interaction/decay length in Air and decides if
        the particle has resonance and/or hadron behavior.

        Class attributes :attr:`is_mixed`, :attr:`E_mix`, :attr:`mix_idx`,
        :attr:`is_resonance` are calculated here. If the option `no_mixing`
        is set in config.adv_config particle is forced to be a resonance
        or hadron behavior.

        Args:
          e_grid (:func:`numpy.array`): energy grid of size :attr:`d`
          max_density (float): maximum density on the integration path (largest
                               decay length)
        """

        info(10, "Calculating mixing energy for", self.name)
        cross_over = config.hybrid_crossover
        max_density = config.max_density

        d = self._energy_grid.d
        inv_intlen = self.inverse_interaction_length()

        inv_declen = self.inverse_decay_length()
        # If particle is stable, no "mixing" necessary
        if (
            not np.any(np.nan_to_num(inv_declen) > 0.0)
            or abs(self.pdg_id[0]) in config.adv_set["exclude_from_mixing"]
            or config.adv_set["no_mixing"]
            or self.pdg_id[0] in config.adv_set["disable_decays"]
        ) and self.is_mixed is not True:
            self.mix_idx = 0
            self.is_mixed = False
            self.is_resonance = False
            return

        # If particle is forced to be a "resonance" or is not in standard_particles
        pid = np.abs(self.pdg_id[0])
        if (
            pid in config.adv_set["force_resonance"]
            and pid not in config.standard_particles
        ):
            self.mix_idx = d
            self.E_mix = self._energy_grid.c[self.mix_idx - 1]
            self.is_mixed = False
            self.is_resonance = True
        # Particle can interact and decay
        elif self.can_interact and not self.is_stable:
            # This is lambda_dec / lambda_int
            threshold = np.zeros_like(inv_intlen)
            mask = inv_declen != 0.0
            threshold[mask] = inv_intlen[mask] * max_density / inv_declen[mask]
            del mask
            mix_idx_check = np.where(threshold >= cross_over)[0]
            if len(mix_idx_check) != 0:
                self.mix_idx = mix_idx_check[0]
                self.E_mix = self._energy_grid.c[self.mix_idx]
            else:
                self.mix_idx = np.where(threshold < cross_over)[0][-1] + 1
                self.E_mix = self._energy_grid.c[self.mix_idx - 1]
            if self.mix_idx != 0:
                self.is_mixed = True
            self.is_resonance = False
        # These particles don't interact but can decay (e.g. tau leptons)
        elif not self.can_interact and not self.is_stable:
            mask = inv_declen != 0.0
            self.mix_idx = np.where(max_density / inv_declen > config.dXmax)[0][0]
            self.E_mix = self._energy_grid.c[self.mix_idx]
            self.is_mixed = True
            self.is_resonance = False
        # Particle is stable but that should be handled above
        else:
            info(0, "This case shouldn't occur. Particle =", self.name)
            threshold = np.inf
            self.mix_idx = 0
            self.is_mixed = False
            self.is_resonance = False

    def __eq__(self, other):
        """Checks name for equality"""
        if isinstance(other, MCEqParticle):
            return self.name == other.name
        return NotImplemented

    def __neq__(self, other):
        """Checks name for equality"""
        if isinstance(other, MCEqParticle):
            return self.name != other.name
        return NotImplemented

    def __hash__(self):
        """Instruction for comuting the hash"""
        return hash(self.name)

    def __repr__(self):
        a_string = f"""
        {self.name}:
        is_hadron     : {self.is_hadron}
        is_lepton     : {self.is_lepton}
        is_nucleus    : {self.is_nucleus}
        is_stable     : {self.is_stable}
        is_mixed      : {self.is_mixed}
        is_resonance  : {self.is_resonance}
        is_tracking   : {self.is_tracking}
        is_projectile : {self.is_projectile}
        mceqidx       : {self.mceqidx}
        E_mix         : {self.E_mix:2.1e} GeV\n"""
        return a_string


class ParticleManager:
    """Database for objects of :class:`MCEqParticle`.

    Authors:
        Anatoli Fedynitch (DESY)
        Jonas Heinze (DESY)
    """

    def __init__(
        self, pdg_id_list, energy_grid, cs_db, medium=config.interaction_medium
    ):
        # (dict) Dimension of primary grid
        self._energy_grid = energy_grid
        # Particle index shortcuts
        #: (dict) Converts PDG ID to index in state vector
        self.pdg2mceqidx = {}
        #: (dict) Converts particle name to index in state vector
        self.pname2mceqidx = {}
        #: (dict) Converts PDG ID to reference of
        # :class:`particlemanager.MCEqParticle`
        self.pdg2pref = {}
        #: (dict) Converts particle name to reference of
        #: class:`particlemanager.MCEqParticle`
        self.pname2pref = {}
        #: (dict) Converts MCEq index to reference of
        #: class:`particlemanager.MCEqParticle`
        self.mceqidx2pref = {}
        #: (dict) Converts index in state vector to PDG ID
        self.mceqidx2pdg = {}
        #: (dict) Converts index in state vector to reference
        # of :class:`particlemanager.MCEqParticle`
        self.mceqidx2pname = {}
        # Save setup of tracked particles to reapply the relations
        # when models change
        self.tracking_relations = {}
        #: (int) Total number of species
        self.nspec = 0
        # save currently applied cross section model
        self.current_cross_sections = None
        # save currently applied hadronic model
        self.current_hadronic_model = None
        # Cross section database
        self._cs_db = cs_db
        # Medium
        self._medium = medium
        # Dictionary to save te tracking particle config
        self.tracking_relations = []
        # Save the tracking relations requested by default tracking
        self._tracking_requested = []

        self._init_categories(particle_pdg_list=pdg_id_list)

        self.print_particle_tables(10)

    def set_cross_sections_db(self, cs_db):
        """Sets the inelastic cross section to each interacting particle.

        This applies to most of the hadrons and does not imply that the
        particle becomes a projectile. parents need in addition defined
        hadronic channels.
        """

        info(5, "Setting cross section particle variables.")
        if self.current_cross_sections == cs_db.iam:
            info(10, "Same cross section model not applied to particles.")
            return

        for p in self.cascade_particles:
            p.set_cs(cs_db)
        self.current_cross_sections = cs_db.iam
        self._update_particle_tables()

    def set_decay_channels(self, decay_db):
        """Attaches the references to the decay yield tables to
        each unstable particle"""

        info(5, "Setting decay info for particles.")
        for p in self.all_particles:
            p.set_decay_channels(decay_db, self)

        self._restore_tracking_setup()
        self._update_particle_tables()

    def set_interaction_model(
        self, cs_db, hadronic_db, updated_parent_list=None, force=False
    ):
        """Attaches the references to the hadronic yield tables to
        each projectile particle"""

        info(5, "Setting hadronic secondaries for particles.")
        if (
            self.current_hadronic_model == hadronic_db.iam
            and not force
            and updated_parent_list is None
        ):
            info(10, "Same hadronic model not applied to particles.")
            return
        if updated_parent_list is not None:
            self._init_categories(updated_parent_list)

        for p in self.cascade_particles:
            p.set_cs(cs_db)
            p.set_hadronic_channels(hadronic_db, self)

        self.current_hadronic_model = hadronic_db.iam
        self._update_particle_tables()

    def set_continuous_losses(self, contloss_db):
        """Set continuous losses terms to particles with ionization
        and radiation losses."""

        for p in self.cascade_particles:
            if p.pdg_id in contloss_db:
                p.has_contloss = True
                p.dEdX = contloss_db[p.pdg_id]
            elif config.generic_losses_all_charged and p.is_charged and not p.is_em:
                # Stopping power dEdX is almost the same for all charged particles.
                # What changes is gamm*beta. We interpolate the dEdX tables
                # stored for protons to different energy grids.
                # Compute beta*gamma from kinetic energy
                betagamma_p = (
                    np.sqrt((self._energy_grid.c + p.mass) ** 2 - p.mass**2) / p.mass
                )
                p.dEdX = -np.exp(contloss_db.generic_spl(np.log(betagamma_p)))
                p.has_contloss = True

    def add_tracking_particle(
        self,
        parent_list,
        child_pdg,
        alias_name,
        from_interactions=False,
        include_antiparticle=False,
    ):
        """Allows tracking decay and particle production chains.

        Replaces previous ``obs_particle`` function that allowed to track
        only leptons from decays certain particles. This present feature
        removes the special PDG IDs 71XX, 72XX, etc and allows to define
        any channel like::

            $ particleManagerInstance.add_tracking_particle([211], 14, 'pi_numu')

        This will store muon neutrinos from pion decays under the alias 'pi_numu'.
        Multiple parents are allowed::

            $ particleManagerInstance.add_tracking_particle(
                [411, 421, 431], 14, 'D_numu')

        Args:

            alias (str): Name alias under which the result is accessible in get_solution
            parents (list): list of parent particle PDG ID's
            child (int): Child particle
            from_interactions (bool): track particles from interactions
        """
        from copy import copy

        info(10, "requested for", parent_list, child_pdg, alias_name)

        for p in parent_list:
            if (
                p,
                child_pdg,
                alias_name,
                from_interactions,
                include_antiparticle,
            ) in self._tracking_requested:
                continue
            self._tracking_requested.append(
                (p, child_pdg, alias_name, from_interactions, include_antiparticle)
            )

        # Check if tracking particle with the alias not yet defined
        # and create new one of necessary
        if alias_name in self.pname2pref:
            info(15, "Re-using tracking particle", alias_name)
            tracking_particle = self.pname2pref[alias_name]
        elif child_pdg not in self.pdg2pref:
            info(
                15,
                "Tracking child not a available",
                "for this interaction model, skipping.",
            )
            return
        else:
            info(10, "Creating new tracking particle")
            # Copy all preferences of the original particle
            tracking_particle = copy(self.pdg2pref[child_pdg])
            tracking_particle.is_tracking = True
            tracking_particle.name = alias_name
            # Find a unique PDG ID for the new tracking particle
            # print child_pdg[0], int(copysign(1000000, child_pdg[0]))
            unique_child_pdg = (
                child_pdg[0] + int(copysign(1000000, child_pdg[0])),
                tracking_particle.helicity,
            )

            for i in range(100):
                if unique_child_pdg not in list(self.pdg2pref):
                    break
                _tn = tracking_particle.name
                _u = unique_child_pdg
                info(
                    20,
                    f"{i}: trying to find unique_pdg ({_tn}) for {_u}",
                )
                unique_child_pdg = (
                    unique_child_pdg[0] + int(copysign(10000, child_pdg[0])),
                    tracking_particle.helicity,
                )
            tracking_particle.unique_pdg_id = unique_child_pdg
            info(
                10,
                "Creating new tracking particle {0} with unique ID {1}".format(
                    tracking_particle.name, unique_child_pdg
                ),
            )

        # Track if attempt to add the tracking particle succeeded at least once
        track_success = False
        # Include antiparticle if requested
        if include_antiparticle:
            par_list = list(set(parent_list + [(-p, h) for (p, h) in parent_list]))
        else:
            par_list = list(set(parent_list))

        for parent_pdg in par_list:
            if parent_pdg not in self.pdg2pref:
                info(15, f"Parent particle {parent_pdg} does not exist.")
                continue
            if (
                parent_pdg,
                child_pdg,
                alias_name,
                from_interactions,
            ) in self.tracking_relations:
                _tn = tracking_particle.name
                _n = self.pdg2pref[parent_pdg].name
                info(
                    20,
                    f"Tracking of {_tn} from {_n} already activated.",
                )
                continue

            if not from_interactions:
                info(15, "Child {0} tracks decays".format(parent_pdg))
                track_method = self.pdg2pref[parent_pdg].track_decays
            else:
                info(15, "Child {0} tracks interactions".format(parent_pdg))
                track_method = self.pdg2pref[parent_pdg].track_interactions

            # Check if the tracking is successful. If not the particle is not
            # a child of the parent particle
            if track_method(tracking_particle):
                info(15, f"Parent particle {parent_pdg} tracking scheduled.")
                self.tracking_relations.append(
                    (parent_pdg, child_pdg, alias_name, from_interactions)
                )
                track_success = True
        if track_success and tracking_particle.name not in list(self.pname2pref):
            tracking_particle.mceqidx = np.max(list(self.mceqidx2pref)) + 1
            self.all_particles.append(tracking_particle)
            self.cascade_particles.append(tracking_particle)
            self._update_particle_tables()
            info(
                10,
                f"tracking particle {tracking_particle.name} successfully added.",
            )

    def track_leptons_from(
        self,
        parent_pdg_list,
        prefix,
        exclude_em=True,
        from_interactions=False,
        use_helicities=False,
        include_antiparticle=False,
    ):
        """Adds tracking particles for all leptons coming from decays of parents
        in `parent_pdg_list`.

        Args:

            parent_pdg_list (list): PDG IDs of parent particles
                                    that leptons originate from
            prefix (list): prefix for each lepton name that will be
                           identified as coming from these parent particles
            exclude_em (bool, optional): exclude electromagnetic (e+-, gamma)
                from leptons
            from_interactions (bool, optional): track particles coming from
                interactions of parents instead of decays
            use_helicities (bool, optional): ignore leptons with
                non-zero/defined helicity
            include_antiparticle (bool, optional): ex. [211, 321] will be
                automatically converted to [211,-211,321,-321]
        """

        leptons = [
            p
            for p in self.all_particles
            if p.is_lepton and not (p.is_em == exclude_em) and not p.is_tracking
        ]

        for lepton in leptons:
            if not use_helicities and lepton.pdg_id[1] != 0:
                continue
            self.add_tracking_particle(
                parent_pdg_list,
                lepton.pdg_id,
                prefix + lepton.name,
                from_interactions,
                include_antiparticle,
            )

    def _init_categories(self, particle_pdg_list):
        """Determines the list of particles for calculation and
        returns lists of instances of :class:`data.MCEqParticle` .

        The particles which enter this list are those, which have a
        defined index in the SIBYLL 2.3 interaction model. Included are
        most relevant baryons and mesons and some of their high mass states.
        More details about the particles which enter the calculation can
        be found in :mod:`particletools`.

        Returns:
          (tuple of lists of :class:`data.MCEqParticle`): (all particles,
          cascade particles, resonances)
        """
        from MCEq.particlemanager import MCEqParticle

        info(5, "Generating particle list.")

        if particle_pdg_list is not None:
            particles = particle_pdg_list
        else:
            from particletools.tables import SibyllParticleTable

            modtab = SibyllParticleTable()
            particles = modtab.baryons + modtab.mesons + modtab.leptons

        # Remove duplicates
        particles = sorted(list(set(particles)))

        # Initialize particle objects
        particle_list = [
            MCEqParticle(
                pdg,
                hel,
                self._energy_grid,
                self._cs_db,
                A_target=average_A_target(self._medium),
            )
            for pdg, hel in particles
        ]

        # Sort by critical energy (= int_len ~== dec_length ~ int_cs/tau)
        particle_list.sort(key=lambda x: x.E_crit, reverse=False)

        # Cascade particles will "live" on the grid and have an mceqidx assigned
        self.cascade_particles = [p for p in particle_list if not p.is_resonance]

        self.cascade_particles = sorted(
            self.cascade_particles, key=lambda p: abs(p.pdg_id[0])
        )

        # These particles will only exist implicitely and integated out
        self.resonances = [p for p in particle_list if p.is_resonance]

        # Assign an mceqidx (position in state vector) to each explicit particle
        # Resonances will kepp the default mceqidx = -1
        for mceqidx, h in enumerate(self.cascade_particles):
            h.mceqidx = mceqidx

        self.all_particles = self.cascade_particles + self.resonances
        self._update_particle_tables()

    def add_new_particle(self, new_mceq_particle):
        if new_mceq_particle in self.all_particles:
            _n = new_mceq_particle.name
            _p = new_mceq_particle.pdg_id
            info(
                0,
                f"Particle {_n}/{_p} has already been added. Use it.",
            )
            return

        if not new_mceq_particle.is_resonance:
            _n = new_mceq_particle.name
            _p = new_mceq_particle.pdg_id
            info(
                2,
                f"New particle {_n}/{_p} is not a resonance.",
            )
            new_mceq_particle.mceqidx = len(self.cascade_particles)
            self.cascade_particles.append(new_mceq_particle)
        else:
            _n = new_mceq_particle.name
            _p = new_mceq_particle.pdg_id
            info(
                2,
                f"New particle {_n}/{_p} is a resonance.",
            )
            self.resonances.append(new_mceq_particle)

        self.all_particles = self.cascade_particles + self.resonances
        self._update_particle_tables()

    def _update_particle_tables(self):
        """Update internal mapping tables after changes to the particle
        list occur."""

        self.n_cparticles = len(self.cascade_particles)
        self.dim = self._energy_grid.d
        self.dim_states = self._energy_grid.d * self.n_cparticles

        # Clean all dictionaries
        [
            d.clear()
            for d in [
                self.pdg2mceqidx,
                self.pname2mceqidx,
                self.mceqidx2pdg,
                self.mceqidx2pname,
                self.mceqidx2pref,
                self.pdg2pref,
                self.pname2pref,
            ]
        ]

        for p in self.all_particles:
            self.pdg2mceqidx[p.unique_pdg_id] = p.mceqidx
            self.pname2mceqidx[p.name] = p.mceqidx
            self.mceqidx2pdg[p.mceqidx] = p.unique_pdg_id
            self.mceqidx2pname[p.mceqidx] = p.name
            self.mceqidx2pref[p.mceqidx] = p
            self.pdg2pref[p.unique_pdg_id] = p
            self.pname2pref[p.name] = p

        self.print_particle_tables(20)

    def _restore_tracking_setup(self):
        """Restores the setup of tracking particles after model changes."""

        info(10, "Restoring tracking particle setup")

        if not self.tracking_relations and config.enable_default_tracking:
            self._init_default_tracking()
            return

        # Clear tracking_relations for this initialization
        self.tracking_relations = []

        for pid, cid, alias, int_dec, incl_anti in self._tracking_requested:
            if pid not in self.pdg2pref:
                info(15, "Can not restore {0}, since not in particle list.")
                continue
            self.add_tracking_particle([pid], cid, alias, int_dec, incl_anti)

    def _init_default_tracking(self):
        """Add default tracking particles for leptons from pi, K, and mu"""
        # Init default tracking particles
        info(1, "Initializing default tracking categories (pi, K, mu)")
        self._tracking_requested_by_default = []
        for parents, prefix, with_helicity in [
            ([(211, 0)], "pi_", True),
            ([(321, 0)], "k_", True),
            ([(13, -1), (13, 1)], "mulr_", False),
            ([(13, 0)], "mu_h0_", False),
            ([(13, -1), (13, 0), (13, 1)], "mu_", False),
            ([(310, 0), (130, 0)], "K0_", False),
        ]:
            self.track_leptons_from(
                parents,
                prefix,
                exclude_em=True,
                use_helicities=with_helicity,
                include_antiparticle=True,
            )

        # Track prompt leptons
        self.track_leptons_from(
            [p.pdg_id for p in self.all_particles if p.ctau < config.prompt_ctau],
            "prcas_",
            exclude_em=True,
            use_helicities=False,
        )
        # Track leptons from interaction vertices (also prompt)
        self.track_leptons_from(
            [p.pdg_id for p in self.all_particles if p.is_projectile],
            "prres_",
            exclude_em=True,
            from_interactions=True,
            use_helicities=False,
        )

        self.track_leptons_from(
            [p.pdg_id for p in self.all_particles if p.is_em],
            "em_",
            exclude_em=True,
            from_interactions=True,
            use_helicities=False,
        )

    def __contains__(self, pdg_id_or_name):
        """Defines the `in` operator to look for particles"""
        if isinstance(pdg_id_or_name, six.integer_types):
            pdg_id_or_name = (pdg_id_or_name, 0)
        elif isinstance(pdg_id_or_name, six.string_types):
            pdg_id_or_name = (_pdata.pdg_id(pdg_id_or_name), 0)
        return pdg_id_or_name in list(self.pdg2pref)

    def __getitem__(self, pdg_id_or_name):
        """Returns reference to particle object."""
        if isinstance(pdg_id_or_name, tuple):
            return self.pdg2pref[pdg_id_or_name]
        if isinstance(pdg_id_or_name, six.integer_types):
            return self.pdg2pref[(pdg_id_or_name, 0)]
        else:
            try:
                return self.pdg2pref[(_pdata.pdg_id(pdg_id_or_name), 0)]
            except KeyError:
                return self.pname2pref[pdg_id_or_name]

    def keys(self):
        """Returns pdg_ids of all particles"""
        return [p.pdg_id for p in self.all_particles]

    def __repr__(self):
        str_out = ""
        ident = 3 * " "
        for s in self.all_particles:
            str_out += s.name + "\n" + ident
            str_out += "PDG id : " + str(s.pdg_id) + "\n" + ident
            str_out += "MCEq idx : " + str(s.mceqidx) + "\n\n"

        return str_out

    def print_particle_tables(self, min_dbg_lev=2):
        info(min_dbg_lev, "Hadrons and stable particles:", no_caller=True)
        print_in_rows(
            min_dbg_lev,
            [
                p.name
                for p in self.all_particles
                if p.is_hadron and not p.is_resonance and not p.is_mixed
            ],
        )

        info(min_dbg_lev, "\nMixed:", no_caller=True)
        print_in_rows(min_dbg_lev, [p.name for p in self.all_particles if p.is_mixed])

        info(min_dbg_lev, "\nResonances:", no_caller=True)
        print_in_rows(
            min_dbg_lev, [p.name for p in self.all_particles if p.is_resonance]
        )

        info(min_dbg_lev, "\nLeptons:", no_caller=True)
        print_in_rows(
            min_dbg_lev,
            [p.name for p in self.all_particles if p.is_lepton and not p.is_tracking],
        )
        info(min_dbg_lev, "\nTracking:", no_caller=True)
        print_in_rows(
            min_dbg_lev, [p.name for p in self.all_particles if p.is_tracking]
        )

        info(
            min_dbg_lev, "\nTotal number of species:", self.n_cparticles, no_caller=True
        )

        # list particle indices
        if False:
            info(10, "Particle matrix indices:", no_caller=True)
            some_index = 0
            for p in self.cascade_particles:
                for i in range(self._energy_grid.d):
                    info(10, p.name + "_" + str(i), some_index, no_caller=True)
                    some_index += 1



import numpy as np

from MCEq import config
from MCEq.misc import info


def solv_numpy(nsteps, dX, rho_inv, int_m, dec_m, phi, grid_idcs):
    """:mod:`numpy` implementation of forward-euler integration.

    Args:
      nsteps (int): number of integration steps
      dX (:func:`numpy.array` [nsteps]): vector of step-sizes :math:`\\Delta X_i` in g/cm**2
      rho_inv (:func:`numpy.array` [nsteps]): vector of density values :math:`\\frac{1}{\\rho(X_i)}`
      int_m (:func:`numpy.array`): interaction matrix :eq:`int_matrix` in dense or sparse representation
      dec_m (:func:`numpy.array`): decay  matrix :eq:`dec_matrix` in dense or sparse representation
      phi (:func:`numpy.array`): initial state vector :math:`\\Phi(X_0)`
    Returns:
      :func:`numpy.array`: state vector :math:`\\Phi(X_{nsteps})` after integration
    """

    grid_sol = []
    grid_step = 0

    imc = int_m
    dmc = dec_m
    dxc = dX
    ric = rho_inv
    phc = phi.copy()  # Fix: create a copy to avoid modifying the input

    dXaccum = 0.0

    from time import time

    start = time()

    for step in range(nsteps):
        phc += (imc.dot(phc) + dmc.dot(ric[step] * phc)) * dxc[step]

        dXaccum += dxc[step]

        if grid_idcs and grid_step < len(grid_idcs) and grid_idcs[grid_step] == step:
            grid_sol.append(np.copy(phc))
            grid_step += 1

    info(
        2,
        f"Performance: {1e3 * (time() - start) / float(nsteps):6.2f}ms/iteration",
    )

    return phc, np.array(grid_sol)


class CUDASparseContext:
    """This class handles the transfer between CPU and GPU memory, and the calling
    of GPU kernels. Initialized by :class:`MCEq.core.MCEqRun` and used by
    :func:`solv_CUDA_sparse`.
    """

    def __init__(self, int_m, dec_m, device_id=config.cuda_gpu_id):
        if config.cuda_fp_precision == 32:
            self.fl_pr = np.float32
        elif config.cuda_fp_precision == 64:
            self.fl_pr = np.float64
        else:
            raise Exception("CUDASparseContext(): Unknown precision specified.")
        # Setup GPU stuff and upload data to it
        try:
            import cupy as cp
            import cupyx.scipy as cpx

            self.cp = cp
            self.cpx = cpx
        except ImportError:
            raise Exception(
                "solv_CUDA_sparse(): CuPy not installed or not available.\n"
                + "Install with: pip install cupy-cuda12x>=12.0.0\n"
                + "CuPy 12.0+ is required for modern sparse matrix interface compatibility."
            )

        cp.cuda.Device(device_id).use()
        self.set_matrices(int_m, dec_m)

    def set_matrices(self, int_m, dec_m):
        """Upload sparce matrices to GPU memory"""
        self.cu_int_m = self.cpx.sparse.csr_matrix(int_m, dtype=self.fl_pr)
        self.cu_dec_m = self.cpx.sparse.csr_matrix(dec_m, dtype=self.fl_pr)

    def alloc_grid_sol(self, dim, nsols):
        """Allocates memory for intermediate if grid solution requested."""
        self.curr_sol_idx = 0
        self.grid_sol = self.cp.zeros((nsols, dim))

    def dump_sol(self):
        """Saves current solution to a new index in grid solution memory."""
        self.cp.copyto(self.grid_sol[self.curr_sol_idx, :], self.cu_curr_phi)
        self.curr_sol_idx += 1
        # self.grid_sol[self.curr_sol, :] = self.cu_curr_phi

    def get_gridsol(self):
        """Downloads grid solution to main memory."""
        return self.cp.asnumpy(self.grid_sol)

    def set_phi(self, phi):
        """Uploads initial condition to GPU memory."""
        self.cu_curr_phi = self.cp.asarray(phi, dtype=self.fl_pr)

    def get_phi(self):
        """Downloads current solution from GPU memory."""
        return self.cp.asnumpy(self.cu_curr_phi)

    def solve_step(self, rho_inv, dX):
        """Makes one solver step on GPU using cuSparse (BLAS)"""

        # Mimic the exact NumPy implementation:
        # phc += (imc.dot(phc) + dmc.dot(ric[step] * phc)) * dxc[step]

        # Calculate: int_m @ curr_phi + dec_m @ (rho_inv * curr_phi)
        int_result = self.cu_int_m @ self.cu_curr_phi
        dec_result = self.cu_dec_m @ (rho_inv * self.cu_curr_phi)
        delta = int_result + dec_result

        # Apply: curr_phi += delta * dX
        self.cu_curr_phi += delta * dX


def solv_CUDA_sparse(nsteps, dX, rho_inv, context, phi, grid_idcs):
    """`NVIDIA CUDA cuSPARSE <https://developer.nvidia.com/cusparse>`_ implementation
    of forward-euler integration.

    Function requires a working :mod:`accelerate` installation.

    Args:
      nsteps (int): number of integration steps
      dX (:func:`numpy.array` [nsteps]): vector of step-sizes :math:`\\Delta X_i` in g/cm**2
      rho_inv (:func:`numpy.array` [nsteps]): vector of density values :math:`\\frac{1}{\\rho(X_i)}`
      int_m (:func:`numpy.array`): interaction matrix :eq:`int_matrix` in dense or sparse representation
      dec_m (:func:`numpy.array`): decay  matrix :eq:`dec_matrix` in dense or sparse representation
      phi (:func:`numpy.array`): initial state vector :math:`\\Phi(X_0)`
      mu_loss_handler (object): object of type :class:`SemiLagrangianEnergyLosses`
    Returns:
      :func:`numpy.array`: state vector :math:`\\Phi(X_{nsteps})` after integration
    """

    c = context
    c.set_phi(phi)

    grid_step = 0

    from time import time

    start = time()
    if len(grid_idcs) > 0:
        c.alloc_grid_sol(phi.shape[0], len(grid_idcs))

    for step in range(nsteps):
        c.solve_step(rho_inv[step], dX[step])

        if grid_idcs and grid_step < len(grid_idcs) and grid_idcs[grid_step] == step:
            c.dump_sol()
            grid_step += 1

    info(
        2,
        f"Performance: {1e3 * (time() - start) / float(nsteps):6.2f}ms/iteration",
    )

    return c.get_phi(), c.get_gridsol() if len(grid_idcs) > 0 else []


def solv_MKL_sparse(nsteps, dX, rho_inv, int_m, dec_m, phi, grid_idcs):
    # mu_loss_handler):
    """`Intel MKL sparse BLAS
    <https://software.intel.com/en-us/articles/
    intel-mkl-sparse-blas-overview?language=en>`_
    implementation of forward-euler integration.

    Function requires that the path to the MKL runtime library ``libmkl_rt.[so/dylib]``
    defined in the config file.

    Args:
      nsteps (int): number of integration steps
      dX (:func:`numpy.array` [nsteps]): vector of step-sizes :math:`\\Delta X_i` in g/cm**2
      rho_inv (:func:`numpy.array` [nsteps]): vector of density values :math:`\\frac{1}{\\rho(X_i)}`
      int_m (:func:`numpy.array`): interaction matrix :eq:`int_matrix` in dense or sparse representation
      dec_m (:func:`numpy.array`): decay  matrix :eq:`dec_matrix` in dense or sparse representation
      phi (:func:`numpy.array`): initial state vector :math:`\\Phi(X_0)`
      grid_idcs (list): indices at which longitudinal solutions have to be saved.

    Returns:
      :func:`numpy.array`: state vector :math:`\\Phi(X_{nsteps})` after integration
    """

    from ctypes import POINTER, Structure, byref, c_int, c_void_p
    from ctypes import c_double as fl_pr

    from MCEq.config import mkl

    # sparse CSR-matrix x dense vector
    create_csr = mkl.mkl_sparse_d_create_csr
    gemv = mkl.mkl_sparse_d_mv
    gemv_hint = mkl.mkl_sparse_set_mv_hint
    optimize = mkl.mkl_sparse_optimize
    # dense vector + dense vector
    axpy = mkl.cblas_daxpy
    np_fl = config.floatlen

    # Prepare CTYPES pointers for MKL sparse CSR BLAS
    int_m_data = int_m.data.ctypes.data_as(POINTER(fl_pr))
    int_m_ci = int_m.indices.ctypes.data_as(POINTER(c_int))
    int_m_pb = int_m.indptr[:-1].ctypes.data_as(POINTER(c_int))
    int_m_pe = int_m.indptr[1:].ctypes.data_as(POINTER(c_int))

    dec_m_data = dec_m.data.ctypes.data_as(POINTER(fl_pr))
    dec_m_ci = dec_m.indices.ctypes.data_as(POINTER(c_int))
    dec_m_pb = dec_m.indptr[:-1].ctypes.data_as(POINTER(c_int))
    dec_m_pe = dec_m.indptr[1:].ctypes.data_as(POINTER(c_int))

    npphi = np.copy(phi).astype(np_fl)
    phi = npphi.ctypes.data_as(POINTER(fl_pr))
    npdelta_phi = np.zeros_like(npphi)
    delta_phi = npdelta_phi.ctypes.data_as(POINTER(fl_pr))

    m = c_int(int_m.shape[0])
    cdzero = fl_pr(0.0)
    cdone = fl_pr(1.0)
    cizero = c_int(0)
    cione = c_int(1)

    grid_step = 0
    grid_sol = []

    int_m_handle = c_void_p()
    mst = create_csr(
        byref(int_m_handle), cizero, m, m, int_m_pb, int_m_pe, int_m_ci, int_m_data
    )

    assert mst == 0, f"MKL create_csr failed with status {mst} on interaction matrix"

    dec_m_handle = c_void_p()
    mst = create_csr(
        byref(dec_m_handle), cizero, m, m, dec_m_pb, dec_m_pe, dec_m_ci, dec_m_data
    )

    assert mst == 0, f"MKL create_csr failed with status {mst} on decay matrix"

    # hints
    operation = int(10)  # SPARSE_OPERATION_NON_TRANSPOSE

    class MatrixDescr(Structure):
        _fields_ = [
            ("type", c_int),
            ("mode", c_int),
            ("diag", c_int),
        ]

    descr = MatrixDescr()
    descr.type = int(20)  # General matrix
    descr.mode = int(121)  # set but dont care since general matrix
    descr.diag = int(131)  # set but dont care since general matrix

    hint_status = gemv_hint(
        int_m_handle,
        operation,
        descr,
        nsteps,
    )

    assert hint_status == 0, (
        f"MKL gemv_hint failed with status {hint_status} on interaction matrix"
    )

    hint_status = gemv_hint(
        dec_m_handle,
        operation,
        descr,
        nsteps,
    )

    assert hint_status == 0, (
        f"MKL gemv_hint failed with status {hint_status} on decay matrix"
    )

    # add mkl_sparse_set_memory_hint???
    #

    o = optimize(int_m_handle)

    assert o == 0, (
        f"MKL mkl_sparse_optimize failed with status {o} on interaction matrix"
    )

    o = optimize(dec_m_handle)

    assert o == 0, f"MKL mkl_sparse_optimize failed with status {o} on decay matrix"

    from time import time

    start = time()

    for step in range(nsteps):
        # delta_phi = int_m.dot(phi)
        gemv(
            operation,
            cdone,
            int_m_handle,
            descr,
            phi,
            cdzero,
            delta_phi,
        )
        # delta_phi = rho_inv * dec_m.dot(phi) + delta_phi
        gemv(
            operation,
            fl_pr(rho_inv[step]),
            dec_m_handle,
            descr,
            phi,
            cdone,
            delta_phi,
        )
        # phi = delta_phi * dX + phi
        axpy(m, fl_pr(dX[step]), delta_phi, cione, phi, cione)

        if grid_idcs and grid_step < len(grid_idcs) and grid_idcs[grid_step] == step:
            grid_sol.append(np.copy(npphi))
            grid_step += 1

    info(
        2,
        f"Performance: {1e3 * (time() - start) / float(nsteps):6.2f}ms/iteration",
    )

    return npphi, np.asarray(grid_sol)


def solv_spacc_sparse(nsteps, dX, rho_inv, spacc_int_m, spacc_dec_m, phi, grid_idcs):
    # mu_loss_handler):
    """Apple Accelerate (vecLib) implementation.

    Args:
      nsteps (int): number of integration steps
      dX (numpy.array[nsteps]): vector of step-sizes
        :math:`\\Delta X_i` in g/cm**2
      rho_inv (numpy.array[nsteps]): vector of density values
        :math:`\\frac{1}{\\rho(X_i)}`
      int_m (numpy.array): interaction matrix :eq:`int_matrix`
        in dense or sparse representation
      dec_m (numpy.array): decay  matrix :eq:`dec_matrix` in
        dense or sparse representation
      phi (numpy.array): initial state vector :math:`\\Phi(X_0)`
      grid_idcs (list): indices at which longitudinal solutions
        have to be saved.

    Returns:
      numpy.array: state vector :math:`\\Phi(X_{nsteps})` after integration
    """

    from ctypes import POINTER, c_double

    import MCEq.spacc as spacc

    dim_phi = int(phi.shape[0])
    npphi = np.copy(phi)
    phi = npphi.ctypes.data_as(POINTER(c_double))
    npdelta_phi = np.zeros_like(npphi)
    delta_phi = npdelta_phi.ctypes.data_as(POINTER(c_double))

    grid_step = 0
    grid_sol = []

    from time import time

    start = time()

    for step in range(nsteps):
        # delta_phi = int_m.dot(phi)
        npdelta_phi *= 0
        spacc_int_m.gemv_ctargs(1.0, phi, delta_phi)

        # delta_phi = rho_inv * dec_m.dot(phi) + delta_phi
        spacc_dec_m.gemv_ctargs(rho_inv[step], phi, delta_phi)

        # phi = delta_phi * dX + phi
        spacc.daxpy(dim_phi, dX[step], delta_phi, phi)

        # axpy(m, fl_pr(dX[step]), delta_phi, cione, phi, cione)

        if grid_idcs and grid_step < len(grid_idcs) and grid_idcs[grid_step] == step:
            grid_sol.append(np.copy(npphi))
            grid_step += 1

    info(
        2,
        "Performance: {0:6.2f}ms/iteration".format(
            1e3 * (time() - start) / float(nsteps)
        ),
    )

    return npphi, np.asarray(grid_sol)


# # TODO: Debug this and transition to BDF
# def _odepack(dXstep=.1,
#                 initial_depth=0.0,
#                 int_grid=None,
#                 grid_var='X',
#                 *args,
#                 **kwargs):
#     """Solves the transport equations with solvers from ODEPACK.

#     Args:
#         dXstep (float): external step size (adaptive sovlers make more
#           steps internally)
#         initial_depth (float): starting depth in g/cm**2
#         int_grid (list): list of depths at which results are recorded
# grid_var (str): Can be depth `X` or something else (currently only `X`
# supported)

#     """
#     from scipy.integrate import ode
#     ri = self.density_model.r_X2rho

#     if config.enable_muon_energy_loss:
#         raise NotImplementedError(
#             'Energy loss not imlemented for this solver.')

#     # Functional to solve
#     def dPhi_dX(X, phi, *args):
#         return self.int_m.dot(phi) + self.dec_m.dot(ri(X) * phi)

#     # Jacobian doesn't work with sparse matrices, and any precision
#     # or speed advantage disappear if used with dense algebra
#     def jac(X, phi, *args):
#         # print 'jac', X, phi
#         return (self.int_m + self.dec_m * ri(X)).todense()

#     # Initial condition
#     phi0 = np.copy(self.phi0)

#     # Initialize variables
#     grid_sol = []

#     # Setup solver
#     r = ode(dPhi_dX).set_integrator(
#         with_jacobian=False, **config.ode_params)

#     if int_grid is not None:
#         initial_depth = int_grid[0]
#         int_grid = int_grid[1:]
#         max_X = int_grid[-1]
#         grid_sol.append(phi0)

#     else:
#         max_X = self.density_model.max_X

#     info(
#         1,
#         'your X-grid is shorter then the material',
#         condition=max_X < self.density_model.max_X)
#     info(
#         1,
#         'your X-grid exceeds the dimentsions of the material',
#         condition=max_X > self.density_model.max_X)

#     # Initial value
#     r.set_initial_value(phi0, initial_depth)

#     info(
#         2, 'initial depth: {0:3.2e}, maximal depth {1:}'.format(
#             initial_depth, max_X))

#     start = time()
#     if int_grid is None:
#         i = 0
#         while r.successful() and (r.t + dXstep) < max_X - 1:
#             info(5, "Solving at depth X =", r.t, condition=(i % 5000) == 0)
#             r.integrate(r.t + dXstep)
#             i += 1
#         if r.t < max_X:
#             r.integrate(max_X)
#         # Do last step to make sure the rational number max_X is reached
#         r.integrate(max_X)
#     else:
#         for i, Xi in enumerate(int_grid):
#             info(5, 'integrating at X =', Xi, condition=i % 10 == 0)

#             while r.successful() and (r.t + dXstep) < Xi:
#                 r.integrate(r.t + dXstep)

#             # Make sure the integrator arrives at requested step
#             r.integrate(Xi)
#             # Store the solution on grid
#             grid_sol.append(r.y)

#     info(2,
#             'time elapsed during integration: {1} sec'.format(time() - start))

#     self.solution = r.y
#     self.grid_sol = grid_sol
